#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of CASCADe package
#
# Developed within the ExoplANETS-A H2020 program.
#
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2018  Jeroen Bouwman
"""
JWST Observatory and Instruments specific module of the CASCADe package
"""
import os
import collections
import ast
from types import SimpleNamespace
import numpy as np
import astropy.units as u
from astropy.io import fits
from astropy.convolution import Gaussian2DKernel
from astropy.convolution import Gaussian1DKernel
from astropy.stats import sigma_clipped_stats
from tqdm import tqdm
import pathlib

from ...initialize import cascade_configuration
from ...initialize import cascade_default_data_path
from ...data_model import SpectralDataTimeSeries
from ...utilities import find, get_data_from_fits
from ..InstrumentsBaseClasses import ObservatoryBase, InstrumentBase
from ...spectral_extraction import rebin_to_common_wavelength_grid


__all__ = ['JWST', 'JWSTMIRILRS', 'JWSTNIRISS', 'JWSTNIRSPEC', 'JWSTNIRCAM']


def get_jwst_instrument_setup(configuration, default_data_path):
    """
    Retrieve all relevant parameters defining the instrument and data setup

    Parameters
    ----------
    configuration :
        CASCADe configuration
    default_data_path :
       CASCADe default data path

    Returns
    -------
    par : `collections.OrderedDict`
        Dictionary containg all relevant parameters

    Raises
    ------
    ValueError
        If obseervationla parameters are not or incorrect defined an
        error will be raised
    """
    # instrument parameters
    inst_obs_name = configuration.instrument_observatory
    inst_inst_name = configuration.instrument
    inst_filter = configuration.instrument_filter

    try:
        inst_spec_order = \
            ast.literal_eval(cascade_configuration.instrument_spectral_order)
    except AttributeError:
        inst_spec_order = None

    # object parameters
    obj_period = \
        u.Quantity(configuration.object_period).to(u.day)
    obj_period = obj_period.value
    obj_ephemeris = \
        u.Quantity(configuration.object_ephemeris).to(u.day)
    obj_ephemeris = obj_ephemeris.value

    # observation parameters
    obs_type = configuration.observations_type
    obs_mode = configuration.observations_mode
    obs_data = configuration.observations_data
    obs_path = configuration.observations_path

    if not os.path.isabs(obs_path):
        obs_path = os.path.join(default_data_path, obs_path)

    obs_id = configuration.observations_id

    obs_data_product = configuration.observations_data_product
    obs_target_name = configuration.observations_target_name
    obs_has_backgr = ast.literal_eval(configuration.
                                      observations_has_background)

    #processing
    try:
        proc_extend_roi = cascade_configuration.processing_extend_roi
        proc_extend_roi = ast.literal_eval(proc_extend_roi)
    except AttributeError:
        proc_extend_roi = [1.0, 1.0, 1.0, 1.0]
    # cpm
    try:
        cpm_ncut_first_int = \
           configuration.cpm_ncut_first_integrations
        cpm_ncut_first_int = ast.literal_eval(cpm_ncut_first_int)
    except AttributeError:
        cpm_ncut_first_int = 0
    try:
        cpm_ncut_last_int = \
           configuration.cpm_ncut_last_integrations
        cpm_ncut_last_int = ast.literal_eval(cpm_ncut_last_int)
    except AttributeError:
        cpm_ncut_last_int = -1
    try:
        proc_rebin_time = int(ast.literal_eval(
            cascade_configuration.processing_rebin_number_time_steps))
    except AttributeError:
        proc_rebin_time = 1
    try:
        proc_rebin_factor = ast.literal_eval(
            cascade_configuration.processing_rebin_factor_spectral_channels)
    except AttributeError:
        proc_rebin_factor = 1.0
    try:
       proc_auto_adjust_rebin_factor = ast.literal_eval(
           cascade_configuration.processing_auto_adjust_rebin_factor_spectral_channels)
    except AttributeError:
        proc_auto_adjust_rebin_factor = False


    par = collections.OrderedDict(
        inst_obs_name=inst_obs_name,
        inst_inst_name=inst_inst_name,
        inst_filter=inst_filter,
        inst_spec_order=inst_spec_order,
        obj_period=obj_period,
        obj_ephemeris=obj_ephemeris,
        obs_type=obs_type,
        obs_mode=obs_mode,
        obs_data=obs_data,
        obs_path=obs_path,
        obs_id=obs_id,
        obs_data_product=obs_data_product,
        obs_target_name=obs_target_name,
        obs_has_backgr=obs_has_backgr,
        cpm_ncut_first_int=cpm_ncut_first_int,
        cpm_ncut_last_int=cpm_ncut_last_int,
        proc_extend_roi=proc_extend_roi,
        proc_rebin_time=proc_rebin_time,
        proc_rebin_factor=proc_rebin_factor,
        proc_auto_adjust_rebin_factor=proc_auto_adjust_rebin_factor)

    return par


def create_mask_from_dq(dq_cube, bits_not_to_flag):
    """
    Create mask from DQ cube.

    Parameters
    ----------
    dq_cube : 'ndarray' of 'float'
        DESCRIPTION.
    bits_not_to_flag : 'list'
        Bit values not to flag

    Returns
    -------
    mask : TYPE
        DESCRIPTION.

    Note
    ----
    Standard bit values not to flag are 0, 12 and 14.
    Bit valiue 10 (blobs) is not set by default but can be selected not to
    be flagged in case of problem.
    """
    bits_to_flag = []
    for ibit in range(1, 16):
        if ibit not in bits_not_to_flag:
            bits_to_flag.append(ibit)
    all_flag_values = np.unique(dq_cube)
    bit_select = np.zeros_like(all_flag_values, dtype='int')
    for ibit in bits_to_flag:
        bit_select = bit_select + (all_flag_values & (1 << (ibit - 1)))
    bit_select = bit_select.astype('bool')
    mask = np.zeros_like(dq_cube, dtype='bool')
    for iflag in all_flag_values[bit_select]:
        mask = mask | (dq_cube == iflag)
    return mask


def read_x1dints_files(data_files, bits_not_to_flag, first_integration,
                       last_integration, spectral_order=None):
    """
    Read the jwst pipeline product x1dints.

    Parameters
    ----------
    data_files : 'list'
        List of x1d.first files.
    bits_not_to_flag : 'list'
        Bit numbers not to use when creating boolean mask from DQ data.
    first_integration : 'int'
        Index of first integration to be used in fit of the entiere data set.
    last_integration : 'int'
        Index of last integration to use.

    Returns
    -------
    wavelength_data : TYPE
        DESCRIPTION.
    spectral_data : TYPE
        DESCRIPTION.
    uncertainty_spectral_data : TYPE
        DESCRIPTION.
    time_bjd : TYPE
        DESCRIPTION.
    mask : TYPE
        DESCRIPTION.
    all_data_files : TYPE
        DESCRIPTION.

    """
    time_bjd = []
    wavelength_data = []
    spectral_data = []
    uncertainty_spectral_data = []
    dq = []

    all_data_files = []

    for data_file in data_files:
        with fits.open(data_file) as hdu_list:
            fits_header = hdu_list[0].header
            nints_end = fits_header['INTEND']
            nints_start = fits_header['INTSTART']
            nints = nints_end-nints_start+1
            all_data_files += [data_file]*nints

            time_bjd += list(hdu_list['INT_TIMES'].data['int_mid_BJD_TDB'])
            for hdu in hdu_list:
                if not (hdu.name == 'EXTRACT1D'):
                    continue
                if spectral_order is not None:
                    if int(hdu.header['SPORDER']) != spectral_order:
                        continue
                idx = np.argsort(hdu.data['WAVELENGTH'])
                wavelength_data.append(hdu.data['WAVELENGTH'][idx])
                spectral_data.append(hdu.data['FLUX'][idx])
                uncertainty_spectral_data.append(hdu.data['FLUX_ERROR'][idx])
                dq.append(hdu.data['DQ'][idx])

    wavelength_data = np.array(wavelength_data, dtype=float).T
    spectral_data = np.array(spectral_data, dtype=float).T
    uncertainty_spectral_data = np.array(uncertainty_spectral_data, dtype=float).T
    dq = np.array(dq, dtype=int).T
    time_bjd = np.array(time_bjd, dtype=float)
    mask = np.ma.masked_invalid(spectral_data).mask
    mask = mask | create_mask_from_dq(dq, bits_not_to_flag)

    idx = np.argsort(time_bjd)[first_integration:last_integration]
    time_bjd = time_bjd[idx]
    spectral_data = spectral_data[:, idx]
    uncertainty_spectral_data = uncertainty_spectral_data[:, idx]
    wavelength_data = wavelength_data[:, idx]
    mask = mask[:, idx]
    all_data_files = [all_data_files[i] for i in idx]

    return (wavelength_data, spectral_data, uncertainty_spectral_data,
            time_bjd, mask, all_data_files)

def read_position_file(data_file, first_integration, last_integration):
    """
    Read file containing position info of spectral source in time.

    Parameters
    ----------
    data_file : 'str'
        Name of file containing position info.
    first_integration : 'int'
        Index of first integration to be used in fit of the entiere data set.
    last_integration : 'int'
        Index of last integration to use.

    Returns
    -------
    'dict'
        Dictionary containing position information.

    """
    position_list = ["POSITION", "TIME_BJD", "DISP_POS", "ANGLE", "SCALE"]
    header_list = ["MEDPOS", "MPUNIT", "PUNIT","DPUNIT", "AUNIT", "SUNIT"]

    if pathlib.Path(data_file).is_file:
        position_dict = {}
        with fits.open(data_file) as hdu_list:

            for key in header_list:
                try:
                    value = hdu_list['PRIMARY'].header[key]
                    position_dict[key] = value
                except:
                    pass
            for key in position_list:
                try:
                    value = hdu_list[1].data[key]
                    position_dict[key] = value[first_integration:last_integration]
                except:
                    pass
        return position_dict
    else:
        return {}


def read_FWHM_file(data_file, first_integration, last_integration):
    """
    Read file containing position info of FWHM of the disperced light in time.

    Parameters
    ----------
    data_file : 'str'
        Name of file containing position info.
    first_integration : 'int'
        Index of first integration to be used in fit of the entiere data set.
    last_integration : 'int'
        Index of last integration to use.

    Returns
    -------
    'dict'
        Dictionary continging FWHM info.

    """
    position_list = ["FWHM", "TIME_BJD"]
    header_list = ['MEDFWHM']

    if pathlib.Path(data_file).is_file:
        position_dict = {}
        with fits.open(data_file) as hdu_list:

            for key in header_list:
                try:
                    value = hdu_list['PRIMARY'].header[key]
                    position_dict[key] = value
                except:
                    pass
            for key in position_list:
                try:
                    value = hdu_list[1].data[key]
                    position_dict[key] = value[first_integration:last_integration]
                except:
                    pass
        return position_dict
    else:
        return {}


class JWST(ObservatoryBase):
    """
    This observatory class defines the instuments and data handling for the
    spectropgraphs of JWST
    """

    def __init__(self):
        # check if cascade is initialized
        if cascade_configuration.isInitialized:
            # check if model is implemented and pick model
            if (cascade_configuration.instrument in
                self.observatory_instruments.keys()):
                factory = self.observatory_instruments[cascade_configuration.instrument]()
                cascade_configuration.telescope_collecting_area = \
                    self.collecting_area
                self.par = factory.par
                self.data = factory.data
                self.spectral_trace = factory.spectral_trace
                if self.par['obs_has_backgr']:
                    self.data_background = factory.data_background
                self.instrument = factory.name
                self.instrument_calibration = \
                    factory.instrument_calibration
                cascade_configuration.instrument_dispersion_scale = \
                     factory.dispersion_scale
            else:
                raise ValueError("JWST instrument not recognized, \
                                 check your init file for the following \
                                 valid instruments: {}. Aborting loading \
                                 instrument".format((self.observatory_instruments).key()))
        else:
            raise ValueError("CASCADe not initialized, \
                                 aborting loading Observatory")

    @property
    def name(self):
        """Set to 'JWST'"""
        return "JWST"

    @property
    def location(self):
        """Set to 'SPACE'"""
        return "SPACE"

    @property
    def collecting_area(self):
        """
        Size of the collecting area of the telescope.

        Returns
        -------
        25.4 m**2
        """
        return '25.40 m2'

    @property
    def NAIF_ID(self):
        """Set to -170 for JWST"""
        return -170

    @property
    def observatory_instruments(self):
        """Returns {'MIRILRS', 'NIRSPECBOTS'}"""
        return {"MIRILRS":JWSTMIRILRS, "NIRSPECBOTS":JWSTNIRSPEC,
                "NIRISSSOSS":JWSTNIRISS, "NIRCAMLW":JWSTNIRCAM}


class JWSTMIRILRS(InstrumentBase):
    """
    JWST MIRI low resolution spectrograph  class.

    This instrument class defines the properties for the LRS spectrograph,
    which is part of the MIRI intrument of JWST.

    For the instrument and observations the following valid options are
    available:

       - data type : {'SPECTRUM'}
       - observing strategy : {'STARING'}
    """

    __valid_data = {'SPECTRUM'}
    __valid_observing_strategy = {'STARING'}

    def __init__(self):
        self.par = self.get_instrument_setup()
        if self.par['obs_has_backgr']:
            self.data, self.data_background = self.load_data()
        else:
            self.data = self.load_data()
        self.spectral_trace = self.get_spectral_trace()
        self._define_region_of_interest()
        try:
            self.instrument_calibration = self.mirilrs_cal
        except AttributeError:
            self.instrument_calibration = None

    @property
    def name(self):
        """
        Name of the JWST instrument: 'MIRILRS'
        """
        return "MIRILRS"

    @property
    def dispersion_scale(self):
        __all_scales = {'P750L': '343.247 Angstrom'}
        return __all_scales[self.par["inst_filter"]]

    @property
    def detector_gain(self):
        """
        Detector gain of the miri detectors.

        Returns
        -------
        2.75 * u.electron / u.DN
        """
        return 2.75 * u.electron / u.DN


    def load_data(self):
        """
        Load the observations.

        This function loads the JWST/MIRI/LRS data form disk based on the
        parameters defined during the initialization of the TSO object.
        """
        if self.par["obs_data"] == 'SPECTRUM':
            data = self.get_spectra()
            if self.par['obs_has_backgr']:
                data_back = self.get_spectra(is_background=True)
        else:
            raise ValueError("The MIRI/LRS insrtrument can currently only be "
                             "used with the observational data parameter "
                             f"set to {self.__valid_data}")
        if self.par['obs_has_backgr']:
            return data, data_back
        else:
            return data

    def get_instrument_setup(self):
        """
        Retrieve all relevant parameters defining the instrument and data setup

        Returns
        -------
        par : `collections.OrderedDict`
            Dictionary containg all relevant parameters

        Raises
        ------
        ValueError
            If obseervationla parameters are not or incorrect defined an
            error will be raised
        """
        par = get_jwst_instrument_setup(cascade_configuration,
                                        cascade_default_data_path)

        return par

    def get_spectra(self, is_background=False):
        """
        Read the input spectra.

        This function combines all functionallity to read fits files
        containing the (uncalibrated) spectral timeseries, including
        orbital phase and wavelength information

        Parameters
        ----------
        is_background : `bool`
            if `True` the data represents an observaton of the IR background
            to be subtracted of the data of the transit spectroscopy target.

        Returns
        -------
        SpectralTimeSeries : `cascade.data_model.SpectralDataTimeSeries`
            Instance of `SpectralDataTimeSeries` containing all spectroscopic
            data including uncertainties, time, wavelength and bad pixel mask.

        Raises
        ------
        AssertionError, KeyError
            Raises an error if no data is found or if certain expected
            fits keywords are not present in the data files.
        """
        # get data files
        if is_background:
            # obsid = self.par['obs_backgr_id']
            target_name = self.par['obs_backgr_target_name']
        else:
            # obsid = self.par['obs_id']
            target_name = self.par['obs_target_name']

        path_to_files = os.path.join(self.par['obs_path'],
                                     self.par['inst_obs_name'],
                                     self.par['inst_inst_name'],
                                     target_name,
                                     'SPECTRA/')

        data_files = find('*' + self.par['obs_id'] + '*' +
                          self.par['obs_data_product']+'.fits',
                          path_to_files)

        # number of integrations
        nintegrations = len(data_files)
        if nintegrations < 1:
            raise AssertionError("No Timeseries data found in dir " +
                                 path_to_files)

        bits_not_to_flag = []
        (wavelength_data, spectral_data, uncertainty_spectral_data, time_bjd,
         mask, all_data_files) = \
            read_x1dints_files(data_files, bits_not_to_flag,
                               self.par["cpm_ncut_first_int"],
                               self.par["cpm_ncut_last_int"])

        # positon info
        position_file = find('target_trace_position.fits', path_to_files)
        if len(position_file) == 0:
            position = np.ones((len(time_bjd)))
            position_unit = u.pix
            median_position = np.median(position)
        else:
            position_dict = read_position_file(position_file[0],
                                      self.par["cpm_ncut_first_int"],
                                      self.par["cpm_ncut_last_int"])
            position = position_dict['POSITION']
            try:
                median_position = position_dict['MEDPOS']
            except:
                median_position = np.median(position)
                position = position - median_position
            try:
                position_unit =  position_dict['PUNIT']
            except:
                position_unit = u.pix

        # FWHM info
        fwhm_file = find('target_FWHM.fits', path_to_files)
        if len(fwhm_file) == 0:
            fwhm = np.ones((len(time_bjd)))
            fwhm_unit = u.pix
            median_fwhm = np.median(fwhm)
        else:
            fwhm_dict = read_FWHM_file(fwhm_file[0],
                                      self.par["cpm_ncut_first_int"],
                                      self.par["cpm_ncut_last_int"])
            fwhm = fwhm_dict['FWHM']
            try:
                median_fwhm = fwhm_dict['MEDFWHM']
            except:
                median_fwhm = np.median(fwhm)
                fwhm = fwhm - median_fwhm
            try:
                fwhm_unit =  fwhm_dict['FWHMUNIT']
            except:
                fwhm_unit = u.pix


        # orbital phase
        phase = ((time_bjd+2400000.5 - self.par['obj_ephemeris']) /
                 self.par['obj_period'])
        phase = phase - int(np.max(phase))
# BUG FIX
        #if np.max(phase) < 0.0:
        #    phase = phase + 1.0
        #phase = phase - np.rint(phase)
        #if self.par['obs_type'] == 'ECLIPSE':
        #    phase[phase < 0] = phase[phase < 0] + 1.0

        detector_gain = self.detector_gain

        scaling = (wavelength_data/7.5)**4
        scaling_unit = u.dimensionless_unscaled
        spectral_data *= detector_gain.value * scaling
        uncertainty_spectral_data *= detector_gain.value * scaling

        wave_unit = u.micron
        flux_unit = u.electron / u.s

        SpectralTimeSeries = \
            SpectralDataTimeSeries(
                                   wavelength=wavelength_data,
                                   wavelength_unit=wave_unit,
                                   data=spectral_data.data,
                                   data_unit=flux_unit,
                                   uncertainty=uncertainty_spectral_data,
                                   time=phase,
                                   time_unit=u.dimensionless_unscaled,
                                   mask=mask,
                                   time_bjd=time_bjd,
                                   position=position,
                                   position_unit=position_unit,
                                   fwhm=fwhm,
                                   fwhm_unit=fwhm_unit,
                                   scaling=scaling,
                                   scaling_unit=scaling_unit,
                                   isRampFitted=True,
                                   isNodded=False,
                                   target_name=target_name,
                                   dataProduct=self.par['obs_data_product'],
                                   dataFiles=all_data_files
                                   )

        # make sure that the date units are as "standard" as posible
        data_unit = (1.0*SpectralTimeSeries.data_unit).decompose().unit
        SpectralTimeSeries.data_unit = data_unit
        wave_unit = (1.0*SpectralTimeSeries.wavelength_unit).decompose().unit
        SpectralTimeSeries.wavelength_unit = wave_unit

        # To make the as standard as posible, by defaut change to
        # mean nomalized data units and use micron as wavelength unit
        mean_signal, _, _ = \
            sigma_clipped_stats(SpectralTimeSeries.return_masked_array("data"),
                                sigma=3, maxiters=10)
        data_unit = u.Unit(mean_signal*SpectralTimeSeries.data_unit)
        SpectralTimeSeries.data_unit = data_unit
        SpectralTimeSeries.wavelength_unit = u.micron

        SpectralTimeSeries.period = self.par['obj_period']
        SpectralTimeSeries.ephemeris = self.par['obj_ephemeris']

        # # rebin in time
        # if self.par['proc_rebin_time'] != 1:
        #     scanDict = {}
        #     idx_scandir = np.ones(spectral_data.shape[-1], dtype=bool)
        #     scanDict[0] = \
        #             {'nsamples': self.par['proc_rebin_time'],
        #              'nscans': sum(idx_scandir),
        #              'index': idx_scandir}

        #     from cascade.spectral_extraction import combine_scan_samples
        #     SpectralTimeSeries = \
        #         combine_scan_samples(SpectralTimeSeries,
        #                              scanDict, verbose=False)

        # # rebin spectra
        # if self.par['proc_auto_adjust_rebin_factor']:
        #     nrebin =  (spectral_data.shape[0]+10) / spectral_data.shape[1]
        # else:
        #     nrebin=self.par['proc_rebin_factor']
        # if nrebin > 1.0:
        #     SpectralTimeSeries = \
        #         rebin_to_common_wavelength_grid(SpectralTimeSeries, 0,
        #                                         nrebin=nrebin, verbose=False,
        #                                         verboseSaveFile=None)

        self._define_convolution_kernel()

        return SpectralTimeSeries


    def get_spectral_trace(self):
        """Get spectral trace."""
        dim = self.data.data.shape
        wave_pixel_grid = np.arange(dim[0]) * u.pix
        position_pixel_grid = np.zeros_like(wave_pixel_grid)
        spectral_trace = \
            collections.OrderedDict(wavelength_pixel=wave_pixel_grid,
                                    positional_pixel=position_pixel_grid,
                                    wavelength=self.data.wavelength.
                                    data[:, 0])
        return spectral_trace

    def _define_region_of_interest(self):
        """
        Defines region on detector which containes the intended target star.
        """
        dim = self.data.data.shape

        wavelength_min = \
            self.par['proc_extend_roi'][0]*4.4*u.micron
        wavelength_max = \
            self.par['proc_extend_roi'][1]*12.5*u.micron

        trace = self.spectral_trace.copy()
        mask_min = trace['wavelength'] > wavelength_min
        mask_max = trace['wavelength'] < wavelength_max
        mask_not_defined = trace['wavelength'] == 0.
        idx_min = int(np.min(trace['wavelength_pixel'].value[mask_min]))
        idx_max = int(np.max(trace['wavelength_pixel'].value[mask_max]))
        roi = np.zeros((dim[0]), dtype=bool)
        roi[0:idx_min] = True
        roi[idx_max+1:] = True
        roi[mask_not_defined] = True

        try:
            self.mirilrs_cal
        except AttributeError:
            self.mirilrs_cal = SimpleNamespace()
        finally:
            self.mirilrs_cal.roi = roi
        return

    def _define_convolution_kernel(self):
        """
        Define convolution kernel.

        Define the instrument specific convolution kernel which will be used
        in the correction procedure of bad pixels.
        """
        if self.par["obs_data"] == 'SPECTRUM':
            kernel = Gaussian1DKernel(4.0, x_size=19)
        else:
            kernel = Gaussian2DKernel(x_stddev=0.2, y_stddev=4.0,
                                      theta=-0.0092, x_size=5, y_size=19)
        try:
            self.mirilrs_cal
        except AttributeError:
            self.mirilrs_cal = SimpleNamespace()
        finally:
            self.mirilrs_cal.convolution_kernel = kernel
        return


class JWSTNIRSPEC(InstrumentBase):
    """
    NIRSPEC instrument module.

    This instrument class defines the properties for the NIRSPEC intrument
    of JWST.

    For the instrument and observations the following valid options are
    available:

       - data type : {'SPECTRUM'}
       - observing strategy : {'STARING'}
    """
    __valid_data = {'SPECTRUM'}
    __valid_observing_strategy = {'STARING'}

    def __init__(self):
        self.par = self.get_instrument_setup()
        if self.par['obs_has_backgr']:
            self.data, self.data_background = self.load_data()
        else:
            self.data = self.load_data()
        self.spectral_trace = self.get_spectral_trace()
        self._define_region_of_interest()
        try:
            self.instrument_calibration = self.nirspecbots_cal
        except AttributeError:
            self.instrument_calibration = None

    @property
    def name(self):
        """
        Name of the JWST instrument: 'NIRSPECBOTS'
        """
        return "NIRSPECBOTS"

    @property
    def dispersion_scale(self):
        __all_scales = {'PRISM-CLEAR': '121.3 Angstrom',
                        'G395M-F290LP': '17.4 Angstrom'}
        return __all_scales[self.par["inst_filter"]]


    @property
    def detector_gain(self):
        """
        Detector gain of the miri detectors.

        Returns
        -------
        0.996 * u.electron / u.DN
        """
        return 0.996 * u.electron / u.DN

    def load_data(self):
        """
        Load the observations.

        This function loads the NIRSPEC BOTS data form disk based on the
        parameters defined during the initialization of the TSO object.
        """
        if self.par["obs_data"] == 'SPECTRUM':
            data = self.get_spectra()
            if self.par['obs_has_backgr']:
                data_back = self.get_spectra(is_background=True)
        else:
            raise ValueError("The NIRSPEC insrtrument can currently only be "
                             "used with the observational data parameter "
                             f"set to {self.__valid_data}")
        if self.par['obs_has_backgr']:
            return data, data_back
        else:
            return data

    def get_instrument_setup(self):
        """
        Retrieve all relevant parameters defining the instrument and data setup

        Returns
        -------
        par : `collections.OrderedDict`
            Dictionary containg all relevant parameters

        Raises
        ------
        ValueError
            If obseervationla parameters are not or incorrect defined an
            error will be raised
        """
        par = get_jwst_instrument_setup(cascade_configuration,
                                        cascade_default_data_path)

        return par

    def get_spectra(self, is_background=False):
        """
        Read the input spectra.

        This function combines all functionallity to read fits files
        containing the (uncalibrated) spectral timeseries, including
        orbital phase and wavelength information

        Parameters
        ----------
        is_background : `bool`
            if `True` the data represents an observaton of the IR background
            to be subtracted of the data of the transit spectroscopy target.

        Returns
        -------
        SpectralTimeSeries : `cascade.data_model.SpectralDataTimeSeries`
            Instance of `SpectralDataTimeSeries` containing all spectroscopic
            data including uncertainties, time, wavelength and bad pixel mask.

        Raises
        ------
        AssertionError, KeyError
            Raises an error if no data is found or if certain expected
            fits keywords are not present in the data files.
        """
        # get data files
        if is_background:
            # obsid = self.par['obs_backgr_id']
            target_name = self.par['obs_backgr_target_name']
        else:
            # obsid = self.par['obs_id']
            target_name = self.par['obs_target_name']

        path_to_files = os.path.join(self.par['obs_path'],
                                     self.par['inst_obs_name'],
                                     self.par['inst_inst_name'],
                                     target_name,
                                     'SPECTRA/')

        data_files = find('*' + self.par['obs_id'] + '*' +
                          self.par['obs_data_product']+'.fits',
                          path_to_files)

        # number of integrations
        nintegrations = len(data_files)
        if nintegrations < 1:
            raise AssertionError("No Timeseries data found in dir " +
                                 path_to_files)

        bits_not_to_flag = []
        (wavelength_data, spectral_data, uncertainty_spectral_data, time_bjd,
         mask, all_data_files) = \
            read_x1dints_files(data_files, bits_not_to_flag,
                               self.par["cpm_ncut_first_int"],
                               self.par["cpm_ncut_last_int"])

        # positon info
        position_file = find('target_trace_position.fits', path_to_files)
        if len(position_file) == 0:
            position = np.ones((len(time_bjd)))
            position_unit = u.pix
            median_position = np.median(position)
        else:
            position_dict = read_position_file(position_file[0],
                                      self.par["cpm_ncut_first_int"],
                                      self.par["cpm_ncut_last_int"])
            position = position_dict['POSITION']
            try:
                median_position = position_dict['MEDPOS']
            except:
                median_position = np.median(position)
                position = position - median_position
            try:
                position_unit =  position_dict['PUNIT']
            except:
                position_unit = u.pix

        # FWHM info
        fwhm_file = find('target_FWHM.fits', path_to_files)
        if len(fwhm_file) == 0:
            fwhm = np.ones((len(time_bjd)))
            fwhm_unit = u.pix
            median_fwhm = np.median(fwhm)
        else:
            fwhm_dict = read_FWHM_file(fwhm_file[0],
                                      self.par["cpm_ncut_first_int"],
                                      self.par["cpm_ncut_last_int"])
            fwhm = fwhm_dict['FWHM']
            try:
                median_fwhm = fwhm_dict['MEDFWHM']
            except:
                median_fwhm = np.median(fwhm)
                fwhm = fwhm - median_fwhm
            try:
                fwhm_unit =  fwhm_dict['FWHMUNIT']
            except:
                fwhm_unit = u.pix

        # orbital phase
        phase = ((time_bjd+2400000.5 - self.par['obj_ephemeris']) /
                 self.par['obj_period'])
        phase = phase - int(np.max(phase))
#        if np.max(phase) < 0.0:
#            phase = phase + 1.0
#        phase = phase - np.rint(phase)
#        if self.par['obs_type'] == 'ECLIPSE':
#            phase[phase < 0] = phase[phase < 0] + 1.0

        detector_gain = self.detector_gain

        scaling = (wavelength_data/3.0)**4
        scaling_unit = u.dimensionless_unscaled
        spectral_data *= detector_gain.value * scaling
        uncertainty_spectral_data *= detector_gain.value * scaling

        wave_unit = u.micron
        flux_unit = u.electron / u.s

        SpectralTimeSeries = \
            SpectralDataTimeSeries(
                                   wavelength=wavelength_data,
                                   wavelength_unit=wave_unit,
                                   data=spectral_data.data,
                                   data_unit=flux_unit,
                                   uncertainty=uncertainty_spectral_data,
                                   time=phase,
                                   time_unit=u.dimensionless_unscaled,
                                   mask=mask,
                                   time_bjd=time_bjd,
                                   position=position,
                                   position_unit=position_unit,
                                   fwhm=fwhm,
                                   fwhm_unit=fwhm_unit,
                                   scaling=scaling,
                                   scaling_unit=scaling_unit,
                                   isRampFitted=True,
                                   isNodded=False,
                                   target_name=target_name,
                                   dataProduct=self.par['obs_data_product'],
                                   dataFiles=all_data_files
                                   )

        # make sure that the date units are as "standard" as posible
        data_unit = (1.0*SpectralTimeSeries.data_unit).decompose().unit
        SpectralTimeSeries.data_unit = data_unit
        wave_unit = (1.0*SpectralTimeSeries.wavelength_unit).decompose().unit
        SpectralTimeSeries.wavelength_unit = wave_unit

        # To make the as standard as posible, by defaut change to
        # mean nomalized data units and use micron as wavelength unit
        mean_signal, _, _ = \
            sigma_clipped_stats(SpectralTimeSeries.return_masked_array("data"),
                                sigma=3, maxiters=10)
        data_unit = u.Unit(mean_signal*SpectralTimeSeries.data_unit)
        SpectralTimeSeries.data_unit = data_unit
        SpectralTimeSeries.wavelength_unit = u.micron

        SpectralTimeSeries.period = self.par['obj_period']
        SpectralTimeSeries.ephemeris = self.par['obj_ephemeris']

        # # rebin in time
        # if self.par['proc_rebin_time'] != 1:
        #     scanDict = {}
        #     idx_scandir = np.ones(spectral_data.shape[-1], dtype=bool)
        #     scanDict[0] = \
        #             {'nsamples': self.par['proc_rebin_time'],
        #              'nscans': sum(idx_scandir),
        #              'index': idx_scandir}

        #     from cascade.spectral_extraction import combine_scan_samples
        #     SpectralTimeSeries = \
        #         combine_scan_samples(SpectralTimeSeries,
        #                              scanDict, verbose=False)

        # # rebin spectra
        # if self.par['proc_auto_adjust_rebin_factor']:
        #     nrebin =  (spectral_data.shape[0]+10) / spectral_data.shape[1]
        # else:
        #     nrebin=self.par['proc_rebin_factor']
        # if nrebin > 1.0:
        #     SpectralTimeSeries = \
        #         rebin_to_common_wavelength_grid(SpectralTimeSeries, 0,
        #                                         nrebin=nrebin, verbose=False,
        #                                         verboseSaveFile=None)

        self._define_convolution_kernel()

        return SpectralTimeSeries


    def get_spectral_images(self):
        """
        Read spectral images.

        Returns
        -------
        None.

        """
        # get data files
        target_name = self.par['obs_target_name']

        path_to_files = os.path.join(self.par['obs_path'],
                                     self.par['inst_obs_name'],
                                     self.par['inst_inst_name'],
                                     target_name,
                                     'SPECTRAL_IMAGES/')

        data_files = find('*' + self.par['obs_id'] + '*rateints.fits',
                          path_to_files)

        # number of integrations
        nintegrations = len(data_files)
        if nintegrations < 1:
            raise AssertionError("No Timeseries data found in dir " +
                                 path_to_files)
        time_bjd = []

        spectral_data = []
        uncertainty_spectral_data = []
        dq = []

        #all_data_files = []

        for data_file in data_files:
            with fits.open(data_file) as hdu_list:
                fits_header = hdu_list[0].header
                exp_start = fits_header['EXPSTART']
                exp_end = fits_header['EXPEND']

                nints_total = fits_header['nints']

                nints = nints_total
                nints_start = 1

                delta_time = (exp_end - exp_start) / nints_total
                start_time = exp_start + 0.5 * delta_time + 2400000.5

                for i in range(2, nints+2):
                    # EXTNAME = 'EXTRACT1D'
                    time_bjd.append(start_time + ((i-2)+(nints_start-1))*delta_time)

                spectral_data.append(hdu_list['SCI'].data)
                uncertainty_spectral_data.append(hdu_list['ERR'].data)
                dq.append(hdu_list['DQ'].data)

        spectral_data = np.concatenate(spectral_data, axis=0, dtype=float).T
        wavelength_data = self.define_wavelength_calibration(spectral_data.shape)
        uncertainty_spectral_data =  np.concatenate(uncertainty_spectral_data, axis=0, dtype=float).T
        dq = np.concatenate(dq, axis=0, dtype=int)

        all_data_files = ['ERS_{0:0=4d}_rateints.fits'.format(i) for i in range(spectral_data.shape[-1])]

        time_bjd = np.array(time_bjd, dtype=float)
        exp_start = np.array(exp_start, dtype=float)

        spectral_data = np.ma.masked_invalid(spectral_data)

        mask = spectral_data.mask

        # orbital phase
        phase = (time_bjd - self.par['obj_ephemeris']) / self.par['obj_period']
        phase = phase - int(np.max(phase))
        if np.max(phase) < 0.0:
            phase = phase + 1.0
        phase = phase - np.rint(phase)
        if self.par['obs_type'] == 'ECLIPSE':
            phase[phase < 0] = phase[phase < 0] + 1.0

        wave_unit = u.micron
        flux_unit = u.ct/u.s

        SpectralTimeSeries = \
            SpectralDataTimeSeries(
                                   wavelength=wavelength_data,
                                   wavelength_unit=wave_unit,
                                   data=spectral_data.data,
                                   data_unit=flux_unit,
                                   uncertainty=uncertainty_spectral_data,
                                   time=phase,
                                   time_unit=u.dimensionless_unscaled,
                                   mask=mask,
                                   time_bjd=time_bjd,
                                   isRampFitted=True,
                                   isNodded=False,
                                   target_name=target_name,
                                   dataProduct=self.par['obs_data_product'],
                                   dataFiles=all_data_files
                                   )
        # make sure that the date units are as "standard" as posible
        data_unit = (1.0*SpectralTimeSeries.data_unit).decompose().unit
        SpectralTimeSeries.data_unit = data_unit
        wave_unit = (1.0*SpectralTimeSeries.wavelength_unit).decompose().unit
        SpectralTimeSeries.wavelength_unit = wave_unit
        # To make the as standard as posible, by defaut change to
        # mean nomalized data units and use micron as wavelength unit
        mean_signal, _, _ = \
            sigma_clipped_stats(SpectralTimeSeries.return_masked_array("data"),
                                sigma=3, maxiters=10)
        data_unit = u.Unit(mean_signal*SpectralTimeSeries.data_unit)
        SpectralTimeSeries.data_unit = data_unit
        SpectralTimeSeries.wavelength_unit = u.micron

        SpectralTimeSeries.period = self.par['obj_period']
        SpectralTimeSeries.ephemeris = self.par['obj_ephemeris']


        if self.par['obs_has_backgr']:
            SpectralTimeSeriesBackground = self.determine_background(SpectralTimeSeries)
            return SpectralTimeSeries, SpectralTimeSeriesBackground

        return SpectralTimeSeries

    def determine_background(self, SpectralTimeSeries):
        import copy
        SpectralTimeSeriesBackground = copy.deepcopy(SpectralTimeSeries)
        data = SpectralTimeSeriesBackground.data.data.value
        background = np.nanmedian(data[:, [0,1,2,3,4,  25,26,27,28,29,30,31], :], axis=1)
        background = np.repeat(background[:,None, :], 32, axis=1)
        background = background * SpectralTimeSeriesBackground.data_unit
        background = np.ma.array(background, mask=SpectralTimeSeriesBackground.mask)
        SpectralTimeSeriesBackground.data = background
        return SpectralTimeSeriesBackground

    def define_wavelength_calibration(self, data_shape):
        coeff = [-5.49471141e-19,  9.88675530e-16, -7.23417941e-13,
                 2.73519357e-10, -5.53539904e-08,  5.33082106e-06,
                 -1.15990213e-04,  3.92862378e-03, 5.61153919e-01]
        x = np.arange(432)
        wave = np.zeros(data_shape[0])
        wave[30:-50] = np.polyval(coeff, x)
        wave_image = np.repeat(wave[..., None], data_shape[1], axis=1)
        wave_cube = np.repeat(wave_image[..., None], data_shape[2], axis=2)
        return wave_cube


    def get_spectral_trace(self):
        """Get spectral trace."""
        dim = self.data.data.shape
        wave_pixel_grid = np.arange(dim[0]) * u.pix
        position_pixel_grid = np.zeros_like(wave_pixel_grid)
        spectral_trace = \
            collections.OrderedDict(wavelength_pixel=wave_pixel_grid,
                                    positional_pixel=position_pixel_grid,
                                    wavelength=self.data.wavelength.
                                    data[:, 0])
        return spectral_trace

    def _define_region_of_interest(self):
        """
        Defines region on detector which containes the intended target star.
        """
        dim = self.data.data.shape
        ndim = self.data.data.ndim

        roi = np.ones((dim[0:ndim-1]), dtype=bool)

        wavelength_min = 2.85*u.micron
        wavelength_max = 5.15*u.micron

        idx = np.where((self.data.wavelength[..., 0].data > wavelength_min) &
                       (self.data.wavelength[..., 0].data < wavelength_max))
        roi[idx] = False

        try:
            self.nirspecbots_cal
        except AttributeError:
            self.nirspecbots_cal = SimpleNamespace()
        finally:
            self.nirspecbots_cal.roi = roi
        return

    def _define_convolution_kernel(self):
        """
        Define convolution kernel.

        Define the instrument specific convolution kernel which will be used
        in the correction procedure of bad pixels.
        """
        if self.par["obs_data"] == 'SPECTRUM':
            kernel = Gaussian1DKernel(4.0, x_size=19)
        else:
            kernel = Gaussian2DKernel(x_stddev=0.2, y_stddev=4.0,
                                      theta=-0.0092, x_size=5, y_size=19)
        try:
            self.nirspecbots_cal
        except AttributeError:
            self.nirspecbots_cal = SimpleNamespace()
        finally:
            self.nirspecbots_cal.convolution_kernel = kernel
        return


class JWSTNIRISS(InstrumentBase):
    """
    NIRISS instrument module.

    This instrument class defines the properties for the NIRISS intrument
    of JWST.

    For the instrument and observations the following valid options are
    available:

       - data type : {'SPECTRUM'}
       - observing strategy : {'STARING'}
    """
    __valid_data = {'SPECTRUM'}
    __valid_observing_strategy = {'STARING'}

    def __init__(self):
        self.par = self.get_instrument_setup()
        if self.par['obs_has_backgr']:
            self.data, self.data_background = self.load_data()
        else:
            self.data = self.load_data()
        self.spectral_trace = self.get_spectral_trace()
        self._define_region_of_interest()
        try:
            self.instrument_calibration = self.nirisssoss_cal
        except AttributeError:
            self.instrument_calibration = None

    @property
    def name(self):
        """
        Name of the JWST instrument: 'NIRISS'
        """
        return "NIRISS"

    @property
    def dispersion_scale(self):
        __all_scales = {'GR700XD-CLEAR': '24.0 Angstrom'}
        return __all_scales[self.par["inst_filter"]]

    @property
    def detector_gain(self):
        """
        Detector gain of the miri detectors.

        Returns
        -------
        1.61 * u.electron / u.DN
        """
        return 1.61 * u.electron / u.DN

    def load_data(self):
        """
        Load the observations.

        This function loads the NIRISS SOSS data form disk based on the
        parameters defined during the initialization of the TSO object.
        """
        if self.par["obs_data"] == 'SPECTRUM':
            data = self.get_spectra()
            if self.par['obs_has_backgr']:
                data_back = self.get_spectra(is_background=True)
        else:
            raise ValueError("The NIRISS insrtrument can currently only be "
                             "used with the observational data parameter "
                             f"set to {self.__valid_data}")
        if self.par['obs_has_backgr']:
            return data, data_back
        else:
            return data

    def get_instrument_setup(self):
        """
        Retrieve all relevant parameters defining the instrument and data setup

        Returns
        -------
        par : `collections.OrderedDict`
            Dictionary containg all relevant parameters

        Raises
        ------
        ValueError
            If obseervationla parameters are not or incorrect defined an
            error will be raised
        """
        par = get_jwst_instrument_setup(cascade_configuration,
                                        cascade_default_data_path)

        return par

    def get_spectra(self, is_background=False):
        """
        Read the input spectra.

        This function combines all functionallity to read fits files
        containing the (uncalibrated) spectral timeseries, including
        orbital phase and wavelength information

        Parameters
        ----------
        is_background : `bool`
            if `True` the data represents an observaton of the IR background
            to be subtracted of the data of the transit spectroscopy target.

        Returns
        -------
        SpectralTimeSeries : `cascade.data_model.SpectralDataTimeSeries`
            Instance of `SpectralDataTimeSeries` containing all spectroscopic
            data including uncertainties, time, wavelength and bad pixel mask.

        Raises
        ------
        AssertionError, KeyError
            Raises an error if no data is found or if certain expected
            fits keywords are not present in the data files.
        """
        # get data files
        if is_background:
            # obsid = self.par['obs_backgr_id']
            target_name = self.par['obs_backgr_target_name']
        else:
            # obsid = self.par['obs_id']
            target_name = self.par['obs_target_name']

        path_to_files = os.path.join(self.par['obs_path'],
                                     self.par['inst_obs_name'],
                                     self.par['inst_inst_name'],
                                     target_name,
                                     'SPECTRA/')

        data_files = find('*' + self.par['obs_id'] + '*' +
                          self.par['obs_data_product']+'.fits',
                          path_to_files)

        # number of integrations
        nintegrations = len(data_files)
        if nintegrations < 1:
            raise AssertionError("No Timeseries data found in dir " +
                                 path_to_files)

        bits_not_to_flag = []
        (wavelength_data, spectral_data, uncertainty_spectral_data, time_bjd,
         mask, all_data_files) = \
            read_x1dints_files(data_files, bits_not_to_flag,
                               self.par["cpm_ncut_first_int"],
                               self.par["cpm_ncut_last_int"],
                               self.par["inst_spec_order"])

        # positon info
        position_file = find('target_trace_position.fits', path_to_files)
        if len(position_file) == 0:
            position = np.ones((len(time_bjd)))
            position_unit = u.pix
            median_position = np.median(position)
        else:
            position_dict = read_position_file(position_file[0],
                                      self.par["cpm_ncut_first_int"],
                                      self.par["cpm_ncut_last_int"])
            position = position_dict['POSITION']
            try:
                median_position = position_dict['MEDPOS']
            except:
                median_position = np.median(position)
                position = position - median_position
            try:
                position_unit =  position_dict['PUNIT']
            except:
                position_unit = u.pix

        # FWHM info
        fwhm_file = find('target_FWHM.fits', path_to_files)
        if len(fwhm_file) == 0:
            fwhm = np.ones((len(time_bjd)))
            fwhm_unit = u.pix
            median_fwhm = np.median(fwhm)
        else:
            fwhm_dict = read_FWHM_file(fwhm_file[0],
                                      self.par["cpm_ncut_first_int"],
                                      self.par["cpm_ncut_last_int"])
            fwhm = fwhm_dict['FWHM']
            try:
                median_fwhm = fwhm_dict['MEDFWHM']
            except:
                median_fwhm = np.median(fwhm)
                fwhm = fwhm - median_fwhm
            try:
                fwhm_unit =  fwhm_dict['FWHMUNIT']
            except:
                fwhm_unit = u.pix

        # orbital phase
        phase = ((time_bjd+2400000.5 - self.par['obj_ephemeris']) /
                 self.par['obj_period'])
        phase = phase - int(np.max(phase))

        detector_gain = self.detector_gain

        scaling = (wavelength_data/1.0)**2
        scaling_unit = u.dimensionless_unscaled
        spectral_data *= detector_gain.value * scaling
        uncertainty_spectral_data *= detector_gain.value * scaling

        wave_unit = u.micron
        flux_unit = u.electron / u.s

        SpectralTimeSeries = \
            SpectralDataTimeSeries(
                                   wavelength=wavelength_data,
                                   wavelength_unit=wave_unit,
                                   data=spectral_data.data,
                                   data_unit=flux_unit,
                                   uncertainty=uncertainty_spectral_data,
                                   time=phase,
                                   time_unit=u.dimensionless_unscaled,
                                   mask=mask,
                                   time_bjd=time_bjd,
                                   position=position,
                                   position_unit=position_unit,
                                   fwhm=fwhm,
                                   fwhm_unit=fwhm_unit,
                                   scaling=scaling,
                                   scaling_unit=scaling_unit,
                                   isRampFitted=True,
                                   isNodded=False,
                                   target_name=target_name,
                                   dataProduct=self.par['obs_data_product'],
                                   dataFiles=all_data_files
                                   )

        # make sure that the date units are as "standard" as posible
        data_unit = (1.0*SpectralTimeSeries.data_unit).decompose().unit
        SpectralTimeSeries.data_unit = data_unit
        wave_unit = (1.0*SpectralTimeSeries.wavelength_unit).decompose().unit
        SpectralTimeSeries.wavelength_unit = wave_unit

        # To make the as standard as posible, by defaut change to
        # mean nomalized data units and use micron as wavelength unit
        mean_signal, _, _ = \
            sigma_clipped_stats(SpectralTimeSeries.return_masked_array("data"),
                                sigma=3, maxiters=10)
        data_unit = u.Unit(mean_signal*SpectralTimeSeries.data_unit)
        SpectralTimeSeries.data_unit = data_unit
        SpectralTimeSeries.wavelength_unit = u.micron

        SpectralTimeSeries.period = self.par['obj_period']
        SpectralTimeSeries.ephemeris = self.par['obj_ephemeris']

        self._define_convolution_kernel()

        return SpectralTimeSeries

    def get_spectral_images(self):
        """
        Read spectral images.

        Returns
        -------
        None.

        """
        pass

    def get_spectral_trace(self):
        """Get spectral trace."""
        dim = self.data.data.shape
        wave_pixel_grid = np.arange(dim[0]) * u.pix
        position_pixel_grid = np.zeros_like(wave_pixel_grid)
        spectral_trace = \
            collections.OrderedDict(wavelength_pixel=wave_pixel_grid,
                                    positional_pixel=position_pixel_grid,
                                    wavelength=self.data.wavelength.
                                    data[:, 0])
        return spectral_trace

    def _define_region_of_interest(self):
        """
        Defines region on detector which containes the intended target star.
        """
        dim = self.data.data.shape
        ndim = self.data.data.ndim

        roi = np.ones((dim[0:ndim-1]), dtype=bool)

        wavelength_min = 0.86*u.micron
        wavelength_max = 2.80*u.micron

        idx = np.where((self.data.wavelength[..., 0].data > wavelength_min) &
                       (self.data.wavelength[..., 0].data < wavelength_max))
        roi[idx] = False

        try:
            self.nirisssoss_cal
        except AttributeError:
            self.nirisssoss_cal = SimpleNamespace()
        finally:
            self.nirisssoss_cal.roi = roi
        return

    def _define_convolution_kernel(self):
        """
        Define convolution kernel.

        Define the instrument specific convolution kernel which will be used
        in the correction procedure of bad pixels.
        """
        if self.par["obs_data"] == 'SPECTRUM':
            kernel = Gaussian1DKernel(4.0, x_size=19)
        else:
            kernel = Gaussian2DKernel(x_stddev=0.8, y_stddev=4.0,
                                      theta=0.0090, x_size=5, y_size=19)
        try:
            self.nirisssoss_cal
        except AttributeError:
            self.nirisssoss_cal = SimpleNamespace()
        finally:
            self.nirisssoss_cal.convolution_kernel = kernel
        return


class JWSTNIRCAM(InstrumentBase):
    """
    NIRCAM LW.

    This instrument class defines the properties for the LW spectrograph,
    which is part of the NIRCAM instrument of the JWST.

    For the instrument and observations the following valid options are
    available:

       - data type : {'SPECTRUM'}
       - observing strategy : {'STARING'}
    """

    __valid_data = {'SPECTRUM'}
    __valid_observing_strategy = {'STARING'}

    def __init__(self):
        self.par = self.get_instrument_setup()
        if self.par['obs_has_backgr']:
            self.data, self.data_background = self.load_data()
        else:
            self.data = self.load_data()
        self.spectral_trace = self.get_spectral_trace()
        self._define_region_of_interest()
        try:
            self.instrument_calibration = self.nircamlw_cal
        except AttributeError:
            self.instrument_calibration = None

    @property
    def name(self):
        """
        Name of the JWST instrument: 'NIRCAMLW'
        """
        return "NIRCAMLW"

    @property
    def dispersion_scale(self):
        __all_scales = {'LW': '100.0 Angstrom'}
        return __all_scales[self.par["inst_filter"]]

    def load_data(self):
        """
        Load the observations.

        This function loads the JWST/NIRCAM/LW data form disk based on the
        parameters defined during the initialization of the TSO object.
        """
        if self.par["obs_data"] == 'SPECTRUM':
            data = self.get_spectra()
            if self.par['obs_has_backgr']:
                data_back = self.get_spectra(is_background=True)
        else:
            raise ValueError("MIRI/LRS instrument can currently only be used \
                              with observational data parameter \
                              set to 'SPECTRUM'")
        if self.par['obs_has_backgr']:
            return data, data_back
        else:
            return data

    def get_instrument_setup(self):
        """
        Retrieve all relevant parameters defining the instrument and data setup

        Returns
        -------
        par : `collections.OrderedDict`
            Dictionary containg all relevant parameters

        Raises
        ------
        ValueError
            If obseervationla parameters are not or incorrect defined an
            error will be raised
        """
        par = get_jwst_instrument_setup(cascade_configuration,
                                        cascade_default_data_path)

        return par

    def get_spectra(self, is_background=False):
        """
        Read the input spectra.

        This function combines all functionallity to read fits files
        containing the (uncalibrated) spectral timeseries, including
        orbital phase and wavelength information

        Parameters
        ----------
        is_background : `bool`
            if `True` the data represents an observaton of the IR background
            to be subtracted of the data of the transit spectroscopy target.

        Returns
        -------
        SpectralTimeSeries : `cascade.data_model.SpectralDataTimeSeries`
            Instance of `SpectralDataTimeSeries` containing all spectroscopic
            data including uncertainties, time, wavelength and bad pixel mask.

        Raises
        ------
        AssertionError, KeyError
            Raises an error if no data is found or if certain expected
            fits keywords are not present in the data files.
        """
        # get data files
        if is_background:
            # obsid = self.par['obs_backgr_id']
            target_name = self.par['obs_backgr_target_name']
        else:
            # obsid = self.par['obs_id']
            target_name = self.par['obs_target_name']

        path_to_files = os.path.join(self.par['obs_path'],
                                     self.par['inst_obs_name'],
                                     self.par['inst_inst_name'],
                                     target_name,
                                     'SPECTRA/')

        data_files = find('*' + self.par['obs_id'] + '*x1dints.fits',
                          path_to_files)

        # number of integrations
        nintegrations = len(data_files)
        if nintegrations < 1:
            raise AssertionError("No Timeseries data found in dir " +
                                 path_to_files)

        bits_not_to_flag = []
        (wavelength_data, spectral_data, uncertainty_spectral_data, time_bjd,
         mask, all_data_files) = \
            read_x1dints_files(data_files, bits_not_to_flag,
                               self.par["cpm_ncut_first_int"])

        # orbital phase
        phase = (time_bjd - self.par['obj_ephemeris']) / self.par['obj_period']
        phase = phase - int(np.max(phase))
        if np.max(phase) < 0.0:
            phase = phase + 1.0
        phase = phase - np.rint(phase)
        if self.par['obs_type'] == 'ECLIPSE':
            phase[phase < 0] = phase[phase < 0] + 1.0

        wave_unit = u.micron
        flux_unit = u.Jy

        #print(np.any(wavelength_data <=0.0))
       # mask[wavelength_data <= 0.0] = True

       # import matplotlib.pyplot as plt
       # plt.imshow(mask)
       # plt.show()


        mask[wavelength_data>4.08] = True

        SpectralTimeSeries = \
            SpectralDataTimeSeries(
                                   wavelength=wavelength_data,
                                   wavelength_unit=wave_unit,
                                   data=spectral_data.data,
                                   data_unit=flux_unit,
                                   uncertainty=uncertainty_spectral_data,
                                   time=phase,
                                   time_unit=u.dimensionless_unscaled,
                                   mask=mask,
                                   time_bjd=time_bjd,
                                   isRampFitted=True,
                                   isNodded=False,
                                   target_name=target_name,
                                   dataProduct=self.par['obs_data_product'],
                                   dataFiles=all_data_files
                                   )
        # make sure that the date units are as "standard" as posible
        data_unit = (1.0*SpectralTimeSeries.data_unit).decompose().unit
        SpectralTimeSeries.data_unit = data_unit
        wave_unit = (1.0*SpectralTimeSeries.wavelength_unit).decompose().unit
        SpectralTimeSeries.wavelength_unit = wave_unit
        # To make the as standard as posible, by defaut change to
        # mean nomalized data units and use micron as wavelength unit
        mean_signal, _, _ = \
            sigma_clipped_stats(SpectralTimeSeries.return_masked_array("data"),
                                sigma=3, maxiters=10)
        data_unit = u.Unit(mean_signal*SpectralTimeSeries.data_unit)
        SpectralTimeSeries.data_unit = data_unit
        SpectralTimeSeries.wavelength_unit = u.micron

        SpectralTimeSeries.period = self.par['obj_period']
        SpectralTimeSeries.ephemeris = self.par['obj_ephemeris']

        ROI = SpectralTimeSeries.mask[:,0]
        from cascade.spectral_extraction import compressDataset
        SpectralTimeSeries, compressMask = compressDataset(SpectralTimeSeries, ROI)


        #print(SpectralTimeSeries.data.shape)
        #print(SpectralTimeSeries.wavelength.shape)
        #print(np.all(np.isfinite(SpectralTimeSeries.wavelength.data)))
        #print(np.any(SpectralTimeSeries.wavelength.data.value <=0.0))

        #print(SpectralTimeSeries.mask[:,0])
        #print(SpectralTimeSeries.mask[:,1])


        #print(SpectralTimeSeries.data)

        if spectral_data.shape[-1] > 512:
            scanDict = {}
            idx_scandir = np.ones(spectral_data.shape[-1], dtype=bool)
            scanDict[0] = \
                    {'nsamples': 2,
                      'nscans': sum(idx_scandir),
                      'index': idx_scandir}

            from cascade.spectral_extraction import combine_scan_samples
            SpectralTimeSeries = \
                combine_scan_samples(SpectralTimeSeries,
                                      scanDict, verbose=False)

        # print(SpectralTimeSeries.data)

        nrebin =  (SpectralTimeSeries.data.shape[0]+10) / SpectralTimeSeries.data.shape[1]
        print(nrebin)
        if nrebin > 1.0:
            SpectralTimeSeries = \
                rebin_to_common_wavelength_grid(SpectralTimeSeries, 0,
                                                nrebin=nrebin, verbose=False,
                                                verboseSaveFile=None)

        self._define_convolution_kernel()

        return SpectralTimeSeries


    def get_spectral_trace(self):
        """Get spectral trace."""
        dim = self.data.data.shape
        wave_pixel_grid = np.arange(dim[0]) * u.pix
        position_pixel_grid = np.zeros_like(wave_pixel_grid)
        spectral_trace = \
            collections.OrderedDict(wavelength_pixel=wave_pixel_grid,
                                    positional_pixel=position_pixel_grid,
                                    wavelength=self.data.wavelength.
                                    data[:, 0])
        return spectral_trace

    def _define_region_of_interest(self):
        """
        Defines region on detector which containes the intended target star.
        """
        dim = self.data.data.shape
        roi = np.zeros((dim[0]), dtype=bool)

        try:
            self.nircamlw_cal
        except AttributeError:
            self.nircamlw_cal = SimpleNamespace()
        finally:
            self.nircamlw_cal.roi = roi
        return

    def _define_convolution_kernel(self):
        """
        Define convolution kernel.

        Define the instrument specific convolution kernel which will be used
        in the correction procedure of bad pixels.
        """
        if self.par["obs_data"] == 'SPECTRUM':
            kernel = Gaussian1DKernel(4.0, x_size=19)
        else:
            kernel = Gaussian2DKernel(x_stddev=0.2, y_stddev=4.0,
                                      theta=-0.0092, x_size=5, y_size=19)
        try:
            self.nircamlw_cal
        except AttributeError:
            self.nircamlw_cal = SimpleNamespace()
        finally:
            self.nircamlw_cal.convolution_kernel = kernel
        return


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of CASCADe package
#
# Developed within the ExoplANETS-A H2020 program.
#
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2018  Jeroen Bouwman
"""
Observatory and Instruments specific module of the CASCADe package defining the
base classes defining the properties of an Instrument or Observatory class. 
"""

from abc import ABCMeta, abstractmethod, abstractproperty

__all__ = ['ObservatoryBase', 'InstrumentBase']


class ObservatoryBase(metaclass=ABCMeta):
    """
    Observatory base class used to define the basic properties an observatory
    class should have
    """
    @abstractproperty
    def name(self):
        """
        Name of the observatory.
        """
        pass

    @abstractproperty
    def location(self):
        """
        Location of the observatory
        """
        pass

    @abstractproperty
    def NAIF_ID(self):
        """
        NAIF ID of the observatory. With this the location relative to the
        sun and the observed target as a function of time can be determined.
        Needed to calculate BJD time.
        """
        pass

    @abstractproperty
    def observatory_instruments(self):
        """
        The names of the instruments part of the observatory.
        """
        pass


class InstrumentBase(metaclass=ABCMeta):
    """
    Instrument base class used to define the basic properties an instrument
    class should have
    """
    @abstractmethod
    def load_data(self):
        """
        Method which allows to load data.
        """
        pass

    @abstractmethod
    def get_instrument_setup(self):
        """
        Method which gets the specific setup of the used instrument.
        """
        pass

    @abstractproperty
    def name(self):
        """
        Name of the instrument.
        """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of the CASCADe package which has been
# developed within the ExoplANETS-A H2020 program.
#
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2018, 2019, 2020  Jeroen Bouwman
"""
Spitzer Observatory and Instruments specific module of the CASCADe package
"""
import os
import collections
import ast
from types import SimpleNamespace

import numpy as np
from astropy.io import fits
import astropy.units as u
from astropy.io import ascii
from astropy.convolution import Gaussian2DKernel
from astropy.convolution import Gaussian1DKernel
from astropy.stats import sigma_clipped_stats
from scipy import interpolate
from scipy.interpolate import griddata
from skimage.morphology import dilation
from skimage.morphology import square
from tqdm import tqdm

from ...initialize import cascade_configuration
from ...initialize import cascade_default_data_path
from ...initialize import cascade_default_path
from ...data_model import SpectralDataTimeSeries
from ...utilities import find, get_data_from_fits
from ..InstrumentsBaseClasses import ObservatoryBase, InstrumentBase

__all__ = ['Spitzer', 'SpitzerIRS']


class Spitzer(ObservatoryBase):
    """
    Class defining the Spitzer observatory.

    This observatory class defines the instuments and data handling for the
    spectropgraphs of the Spitzer Space telescope
    """

    def __init__(self):
        # check if cascade is initialized
        if cascade_configuration.isInitialized:
            # check if model is implemented and pick model
            if (cascade_configuration.instrument in
                    self.observatory_instruments):
                if cascade_configuration.instrument == 'IRS':
                    factory = SpitzerIRS()
                    self.par = factory.par
                    cascade_configuration.telescope_collecting_area = \
                        self.collecting_area
                    self.data = factory.data
                    self.spectral_trace = factory.spectral_trace
                    if self.par['obs_has_backgr']:
                        self.data_background = factory.data_background
                    self.instrument = factory.name
                    self.instrument_calibration = \
                        factory.instrument_calibration
                    cascade_configuration.instrument_dispersion_scale = \
                        factory.dispersion_scale
            else:
                raise ValueError("Spitzer instrument not recognized, \
                                 check your init file for the following \
                                 valid instruments: {}. Aborting loading \
                                 instrument".format(self.valid_instruments))
        else:
            raise ValueError("CASCADe not initialized, \
                                 aborting loading Observatory")

    @property
    def name(self):
        """Name of the observatory."""
        return "SPITZER"

    @property
    def location(self):
        """Location of the observatory."""
        return "SPACE"

    @property
    def collecting_area(self):
        """
        Size of the collecting area of the telescope.

        Returns
        -------
        0.5 m**2
        """
        return '0.5 m2'

    @property
    def NAIF_ID(self):
        """NAIF_ID of the observatory."""
        return -79

    @property
    def observatory_instruments(self):
        """All implemented instruments of the observatory."""
        return{"IRS"}


class SpitzerIRS(InstrumentBase):
    """
    Class defining the IRS instrument.

    This instrument class defines the properties of the IRS instrument of
    the Spitzer Space Telescope.
    For the instrument and observations the following valid options are
    available:

       - detectors :  {'SL', 'LL'}
       - spectral orders : {'1', '2'}
       - data products : {'droop', 'COE'}
       - observing mode : {'STARING', 'NODDED'}
       - data type : {'SPECTRUM', 'SPECTRAL_IMAGE', 'SPECTRAL_DETECTOR_CUBE'}

    """

    __valid_arrays = {'SL', 'LL'}
    __valid_orders = {'1', '2'}
    __valid_filters = {'SL1', 'SL2', 'LL1', 'LL2'}
    __valid_data = {'SPECTRUM', 'SPECTRAL_IMAGE', 'SPECTRAL_DETECTOR_CUBE'}
    __valid_observing_strategy = {'STARING', 'NODDED'}
    __valid_data_products = {'droop', 'COE', 'FEPS', 'CAE'}

    def __init__(self):

        self.par = self.get_instrument_setup()
        if self.par['obs_has_backgr']:
            self.data, self.data_background = self.load_data()
        else:
            self.data = self.load_data()
        self.spectral_trace = self.get_spectral_trace()
        self._define_region_of_interest()
        try:
            self.instrument_calibration = self.IRS_cal
        except AttributeError:
            self.instrument_calibration = None

    @property
    def name(self):
        """Reteurn instrument name."""
        return "IRS"

    @property
    def dispersion_scale(self):
        __all_scales = {'SL1': '604.5 Angstrom', 'SL2': '371.7 Angstrom',
                        'LL2': '743.4 Angstrom', 'LL1': '1209.0 Angstrom'}
        return __all_scales[self.par["inst_filter"]]

    def load_data(self):
        """Load data."""
        if self.par["obs_data"] == 'SPECTRUM':
            data = self.get_spectra()
            if self.par['obs_has_backgr']:
                data_back = self.get_spectra(is_background=True)
        elif self.par["obs_data"] == 'SPECTRAL_IMAGE':
            data = self.get_spectral_images()
            if self.par['obs_has_backgr']:
                data_back = self.get_spectral_images(is_background=True)
        elif self.par["obs_data"] == 'SPECTRAL_DETECTOR_CUBE':
            data = self.get_detector_cubes()
            if self.par['obs_has_backgr']:
                data_back = self.get_detector_cubes(is_background=True)
        if self.par['obs_has_backgr']:
            return data, data_back
        else:
            return data

    def get_instrument_setup(self):
        """Retrieve all parameters defining the instrument and data setup."""
        inst_obs_name = cascade_configuration.instrument_observatory
        inst_inst_name = cascade_configuration.instrument
        inst_filter = cascade_configuration.instrument_filter
        obj_period = \
            u.Quantity(cascade_configuration.object_period).to(u.day)
        obj_period = obj_period.value
        obj_ephemeris = \
            u.Quantity(cascade_configuration.object_ephemeris).to(u.day)
        obj_ephemeris = obj_ephemeris.value
        obs_mode = cascade_configuration.observations_mode
        obs_data = cascade_configuration.observations_data
        obs_path = cascade_configuration.observations_path
        if not os.path.isabs(obs_path):
            obs_path = os.path.join(cascade_default_data_path, obs_path)
        obs_cal_path = cascade_configuration.observations_cal_path
        if not os.path.isabs(obs_cal_path):
            obs_cal_path = os.path.join(cascade_default_path,
                                        obs_cal_path)
        obs_cal_version = cascade_configuration.observations_cal_version
        obs_data_product = cascade_configuration.observations_data_product
        obs_id = cascade_configuration.observations_id
        obs_target_name = cascade_configuration.observations_target_name
        obs_has_backgr = ast.literal_eval(cascade_configuration.
                                          observations_has_background)
        if obs_has_backgr:
            obs_backgr_id = cascade_configuration.observations_background_id
            obs_backgr_target_name = \
                cascade_configuration.observations_background_name

        # cpm
        try:
            cpm_ncut_first_int = \
               cascade_configuration.cpm_ncut_first_integrations
            cpm_ncut_first_int = ast.literal_eval(cpm_ncut_first_int)
        except AttributeError:
            cpm_ncut_first_int = 0

        if not (obs_data in self.__valid_data):
            raise ValueError("Data type not recognized, "
                             "check your init file for the following "
                             "valid types: {}. "
                             "Aborting loading data".format(self.__valid_data))
        if not (obs_mode in self.__valid_observing_strategy):
            raise ValueError("Observational stategy not recognized, "
                             "check your init file for the following "
                             "valid types: {}. Aborting loading "
                             "data".format(self.__valid_observing_strategy))
        if not (inst_filter in self.__valid_filters):
            raise ValueError("Spectral filter not recognized, "
                             "check your init file for the following "
                             "valid types: {}. Aborting loading "
                             "data".format(self.__valid_filters))
        inst_mode = inst_filter[0:2]
        inst_order = inst_filter[2]
        if not (inst_mode in self.__valid_arrays):
            raise ValueError("Instrument mode not recognized, "
                             "check your init file for the following "
                             "valid types: {}. Aborting "
                             "loading data".format(self.__valid_arrays))
        if not (inst_order in self.__valid_orders):
            raise ValueError("Spectral order not recognized, "
                             "check your init file for the following "
                             "valid types: {}. Aborting loading "
                             "data".format(self.__valid_orders))
        if not (obs_data_product in self.__valid_data_products):
            raise ValueError("Data product not recognized, "
                             "check your init file for the following "
                             "valid types: {}. Aborting loading "
                             "data".format(self.__valid_data_products))
        par = collections.OrderedDict(inst_obs_name=inst_obs_name,
                                      inst_inst_name=inst_inst_name,
                                      inst_filter=inst_filter,
                                      inst_mode=inst_mode,
                                      inst_order=inst_order,
                                      obj_period=obj_period,
                                      obj_ephemeris=obj_ephemeris,
                                      obs_mode=obs_mode,
                                      obs_data=obs_data,
                                      obs_path=obs_path,
                                      obs_cal_path=obs_cal_path,
                                      obs_data_product=obs_data_product,
                                      obs_cal_version=obs_cal_version,
                                      obs_id=obs_id,
                                      obs_target_name=obs_target_name,
                                      obs_has_backgr=obs_has_backgr,
                                      cpm_ncut_first_int=cpm_ncut_first_int)
        if obs_has_backgr:
            par.update({'obs_backgr_id': obs_backgr_id})
            par.update({'obs_backgr_target_name': obs_backgr_target_name})
        return par

    def get_spectra(self, is_background=False):
        """
        Get the 1D spectra.

        This function combines all functionallity to read fits files
        containing the (uncalibrated) spectral timeseries, including
        orbital phase and wavelength information

        Parameters
        ----------
        is_background : `bool`
            if `True` the data represents an observaton of the IR background
            to be subtracted of the data of the transit spectroscopy target.

        Returns
        -------
        SpectralTimeSeries : `cascade.data_model.SpectralDataTimeSeries`
            Instance of `SpectralDataTimeSeries` containing all spectroscopic
            data including uncertainties, time, wavelength and bad pixel mask.

        Raises
        ------
        AssertionError, KeyError
            Raises an error if no data is found or if certain expected
            fits keywords are not present in the data files.
        """
        # get data files
        if is_background:
            target_name = self.par['obs_backgr_target_name']
        else:
            target_name = self.par['obs_target_name']

        path_to_files = os.path.join(self.par['obs_path'],
                                     self.par['inst_obs_name'],
                                     self.par['inst_inst_name'],
                                     target_name,
                                     'SPECTRA/')
        data_files = find('*' + self.par['obs_id'] + '*' +
                          self.par['obs_data_product']+'.fits', path_to_files)

        # number of integrations
        nintegrations = len(data_files)
        if nintegrations < 2:
            raise AssertionError("No Timeseries data found in dir " +
                                 path_to_files)

        data_list = ['LAMBDA', 'FLUX', 'FERROR', 'MASK']
        auxilary_list = ["POSITION", "MEDPOS", "PUNIT", "MPUNIT", "TIME_BJD",
                         "DISP_POS", "ANGLE", "SCALE", "DPUNIT", "AUNIT",
                         "SUNIT"]

        data_dict, auxilary_dict = \
            get_data_from_fits(data_files, data_list, auxilary_list)

        if (not auxilary_dict['TIME_BJD']['flag']):
            raise KeyError("No TIME_BJD keyword found in fits files")

        wavelength_data = np.array(data_dict['LAMBDA']['data']).T
        wave_unit = data_dict['LAMBDA']['data'][0].unit
        spectral_data = np.array(data_dict['FLUX']['data']).T
        flux_unit = data_dict['FLUX']['data'][0].unit
        uncertainty_spectral_data = np.array(data_dict['FERROR']['data']).T
        if data_dict['MASK']['flag']:
            mask = np.array(data_dict['MASK']['data']).T
        else:
            mask = np.zeros_like(spectral_data, dtype=bool)
        if auxilary_dict['TIME_BJD']['flag']:
            time = np.array(auxilary_dict['TIME_BJD']['data']) * u.day
            phase = (time.value - self.par['obj_ephemeris']) / \
                self.par['obj_period']
            phase = phase - int(np.max(phase))
            if np.max(phase) < 0.0:
                phase = phase + 1.0

        position = np.array(auxilary_dict['POSITION']['data'])
        posUnit =  auxilary_dict['PUNIT']['data_unit']
        angle =  np.array(auxilary_dict['ANGLE']['data'])
        angleUnit = auxilary_dict['AUNIT']['data_unit']
        scaling = np.array(auxilary_dict['SCALE']['data'])
        scaleUnit = auxilary_dict['SUNIT']['data_unit']
        dispersion_position = np.array(auxilary_dict['DISP_POS']['data'])
        dispPosUnit = auxilary_dict['DPUNIT']['data_unit']

        idx = np.argsort(time)[self.par["cpm_ncut_first_int"]:]
        time = time[idx]
        spectral_data = spectral_data[:, idx]
        uncertainty_spectral_data = uncertainty_spectral_data[:, idx]
        wavelength_data = wavelength_data[:, idx]
        mask = mask[:, idx]
        data_files = [data_files[i] for i in idx]
        phase = phase[idx]
        position = position[idx]
        angle = angle[idx]
        scaling = scaling[idx]
        dispersion_position = dispersion_position[idx]

        SpectralTimeSeries = \
            SpectralDataTimeSeries(wavelength=wavelength_data,
                                   wavelength_unit=wave_unit,
                                   data=spectral_data,
                                   data_unit=flux_unit,
                                   uncertainty=uncertainty_spectral_data,
                                   time=phase,
                                   time_unit=u.dimensionless_unscaled,
                                   mask=mask,
                                   time_bjd=time,
                                   position=position,
                                   position_unit=posUnit,
                                   isRampFitted=True,
                                   isNodded=False,
                                   target_name=target_name,
                                   dataProduct=self.par['obs_data_product'],
                                   dataFiles=data_files)
        # Standardize signal to mean value.
        mean_signal, _, _ = \
            sigma_clipped_stats(SpectralTimeSeries.return_masked_array("data"),
                                sigma=3, maxiters=10)
        data_unit = u.Unit(mean_signal*SpectralTimeSeries.data_unit)
        SpectralTimeSeries.data_unit = data_unit
        SpectralTimeSeries.wavelength_unit = u.micron
        SpectralTimeSeries.add_measurement(
            disp_position=dispersion_position,
            disp_position_unit=dispPosUnit,
            angle=angle,
            angle_unit=angleUnit,
            scale=scaling,
            scale_unit=scaleUnit)

        self._define_convolution_kernel()

        return SpectralTimeSeries

    def get_spectral_images(self, is_background=False):
        """
        Get the 2D spectral images.

        This function combines all functionallity to read fits files
        containing the (uncalibrated) spectral timeseries, including
        orbital phase and wavelength information

        Parameters
        ----------
        is_background : `bool`
            if `True` the data represents an observaton of the IR background
            to be subtracted of the data of the transit spectroscopy target.

        Returns
        -------
        SpectralTimeSeries : `cascade.data_model.SpectralDataTimeSeries`
            Instance of `SpectralDataTimeSeries` containing all spectroscopic
            data including uncertainties, time, wavelength and bad pixel mask.

        Raises
        ------
        AssertionError, KeyError
            Raises an error if no data is found or if certain expected
            fits keywords are not present in the data files.

        Notes
        -----
        Notes on FOV:
            in the fits header the following relevant info is used:

            - FOVID     26     IRS_Short-Lo_1st_Order_1st_Position
            - FOVID     27     IRS_Short-Lo_1st_Order_2nd_Position
            - FOVID     28     IRS_Short-Lo_1st_Order_Center_Position
            - FOVID     29     IRS_Short-Lo_Module_Center
            - FOVID     32     IRS_Short-Lo_2nd_Order_1st_Position
            - FOVID     33     IRS_Short-Lo_2nd_Order_2nd_Position
            - FOVID     34     IRS_Short-Lo_2nd_Order_Center_Position
            - FOVID     40     IRS_Long-Lo_1st_Order_Center_Position
            - FOVID     46     IRS_Long-Lo_2nd_Order_Center_Position

        Notes on timing:

            - FRAMTIME the total effective exposure time (ramp length)
              in seconds

        """
        # order mask
        mask = self._get_order_mask()

        # get data files
        if is_background:
            obsid = self.par['obs_backgr_id']
            target_name = self.par['obs_backgr_target_name']
        else:
            obsid = self.par['obs_id']
            target_name = self.par['obs_target_name']
        path_to_files = os.path.join(self.par['obs_path'],
                                     self.par['inst_obs_name'],
                                     self.par['inst_inst_name'],
                                     target_name,
                                     'SPECTRAL_IMAGES/')
        data_files = find('*' + obsid + '*_' +
                          self.par['obs_data_product']+'.fits', path_to_files)

        # number of integrations
        nintegrations = len(data_files)
        if nintegrations < 2:
            raise AssertionError("No Timeseries data found in dir " +
                                 path_to_files)

        # read in the first image and get information on the observations
        # get the size of the spectral images
        image_file = data_files[0]
        spectral_image = fits.getdata(image_file, ext=0)
        npix, mpix = spectral_image.shape
        # get frametime from fits header (duration of ramp in sec)
        framtime = fits.getval(image_file, "FRAMTIME", ext=0)
        # get the FOVID from the header and check for nodded observations
        try:
            fovid = fits.getval(image_file, "FOVID", ext=0)
        except KeyError:
            print("FOVID not set in fits files")
            print("Using 'observations_mode' parameter instead")
            if self.par['obs_mode'] == "STARING":
                fovid = 28
            else:
                fovid = 26
        if (fovid != 28) and (fovid != 34) and (fovid != 40) and (fovid != 46):
            isNodded = True
            nodOffset = -5.0 * u.pix
        else:
            isNodded = False
            nodOffset = 0.0 * u.pix

        # get the unit of the spectral images
        try:
            flux_unit_string = fits.getval(image_file, "BUNIT", ext=0)
            flux_unit_string = flux_unit_string.replace("e-", "electron")
            flux_unit_string = flux_unit_string.replace("sec", "s")
            flux_unit = u.Unit(flux_unit_string)
        except KeyError:
            print("No flux unit set in fits files")
            flux_unit = u.dimensionless_unscaled

        # define mask and fill with data  from order mask
        mask = np.tile(mask.T, (nintegrations, 1, 1)).T

        # get the data
        image_cube = np.zeros((npix, mpix, nintegrations))
        image_unc_cube = np.zeros((npix, mpix, nintegrations))
        image_dq_cube = np.zeros((npix, mpix, nintegrations))
        time = np.zeros((nintegrations))
        for im, image_file in enumerate(tqdm(data_files, dynamic_ncols=True)):
            # WARNING fits data is single precision!!
            spectral_image = fits.getdata(image_file, ext=0)
            image_cube[:, :, im] = spectral_image
            time[im] = fits.getval(image_file, "BMJD_OBS", ext=0)
            unc_image = fits.getdata(image_file.replace("droop.fits",
                                                        "drunc.fits"), ext=0)
            image_unc_cube[:, :, im] = unc_image
            dq_image = fits.getdata(image_file.replace("droop.fits",
                                                       "bmask.fits"), ext=0)
            image_dq_cube[:, :, im] = dq_image

        data = image_cube * flux_unit
        uncertainty = image_unc_cube * flux_unit
        npix, mpix, nintegrations = data.shape

        mask_dq = image_dq_cube > int('10000000', 2)
        mask = np.logical_or(mask, mask_dq)

        # The time in the spitzer fits header is -2400000.5 and
        # it is the time at the start of the ramp
        # As we are using fitted ramps,
        # shift time by half ramp of length framtime
        time = (time + 2400000.5) + (0.50*framtime) / (24.0*3600.0)  # in days
        # orbital phase
        phase = (time - self.par['obj_ephemeris']) / self.par['obj_period']
        phase = phase - np.int(np.max(phase))
        if np.max(phase) < 0.0:
            phase = phase + 1.0

        self._define_convolution_kernel()

        # wavelength calibration
        wave_cal = self._get_wavelength_calibration(npix, nodOffset)

        # reverse spectral data to make sure the shortest wavelengths are
        # at the first image row
        SpectralTimeSeries = \
            SpectralDataTimeSeries(wavelength=wave_cal[::-1, ...],
                                   data=data[::-1, ...],
                                   time=phase,
                                   uncertainty=uncertainty[::-1, ...],
                                   mask=mask[::-1, ...],
                                   time_bjd=time,
                                   isRampFitted=True,
                                   isNodded=isNodded,
                                   target_name=target_name,
                                   dataProduct=self.par['obs_data_product'],
                                   dataFiles=data_files)
        return SpectralTimeSeries

    def _define_convolution_kernel(self):
        """
        Define the instrument specific convolution kernel.

        This function defines an instrument specific convolution kernel
        which will be used in the correction procedure of bad pixels.
        """
        if self.par["obs_data"] == 'SPECTRUM':
            kernel = Gaussian1DKernel(2.2, x_size=13)
        else:
            kernel = Gaussian2DKernel(x_stddev=0.2, y_stddev=2.2, theta=0.076,
                                      x_size=5, y_size=13)
        try:
            self.IRS_cal
        except AttributeError:
            self.IRS_cal = SimpleNamespace()
        finally:
            self.IRS_cal.convolution_kernel = kernel
        return

    def _define_region_of_interest(self):
        """
        Define region on detector.

        This functon defines the region of interest which containes
        the intended target star. It defines a mask such that all data flagged
        or of no interest for the data calibraiton and spectral extraction.
        """
        dim = self.data.data.shape
        if len(dim) <= 2:
            roi = np.zeros((dim[0]), dtype=bool)
            roi[0] = True
            roi[-1] = True
        else:
            roi = self._get_order_mask()
            selem = square(1)
            roi = dilation(roi, selem)
            roi[:, 0:1] = True
            roi[:, -1:] = True
            roi = roi[::-1, ...]
        try:
            self.IRS_cal
        except AttributeError:
            self.IRS_cal = SimpleNamespace()
        finally:
            self.IRS_cal.roi = roi
        return

    def _get_order_mask(self):
        """
        Get the order mask.

        This functions gets the mask which defines the pixels used
        with a given spectral order
        """
        # order mask
        order_mask_file_name = \
            os.path.join(self.par['obs_cal_path'],
                         self.par['inst_obs_name'],
                         self.par['inst_inst_name'],
                         self.par['obs_cal_version'],
                         'IRSX_'+self.par['inst_mode']+'_' +
                         self.par['obs_cal_version']+'_cal.omask.fits')
        order_masks = fits.getdata(order_mask_file_name, ext=0)

        if self.par['inst_order'] == '1':
            mask = np.ones(shape=order_masks.shape, dtype=bool)
            mask[order_masks == 1] = False
            # as there are often problems at the edge of the detector array,
            # cut first and last row
            row_check = np.all(mask, axis=1)
            idx_row = np.arange(mask.shape[0])
            # remove last row
            mask[idx_row[np.logical_not(row_check)][-1], :] = True
            # remove first row
            mask[idx_row[np.logical_not(row_check)][0], :] = True
            # remove bad part of LL1
            if self.par['inst_mode'] == 'LL':
                mask[idx_row[np.logical_not(row_check)][:50], :] = True
                mask[idx_row[np.logical_not(row_check)][-5:], :] = True

        elif (self.par['inst_order'] == '2') or \
                (self.par['inst_order'] == '3'):
            # SL2 or LL2
            mask1 = np.ones(shape=order_masks.shape, dtype=bool)
            mask1[order_masks == 2] = False
            # as there are often problems at the edge of the detector array,
            # cut first and last row
            row_check = np.all(mask1, axis=1)
            idx_row = np.arange(mask1.shape[0])
            # remove last row
            mask1[idx_row[np.logical_not(row_check)][-1], :] = True
            # remove first row
            mask1[idx_row[np.logical_not(row_check)][0], :] = True

            if self.par['inst_mode'] == 'LL':
                mask1[idx_row[np.logical_not(row_check)][79:83], :] = True
                mask1[idx_row[np.logical_not(row_check)][:5], :] = True
            # SL3 or LL3
            mask2 = np.ones(shape=order_masks.shape, dtype=bool)
            mask2[order_masks == 3] = False
            # as there are often problems at the edge of the detector array,
            # cut first and last row
            row_check = np.all(mask2, axis=1)
            idx_row = np.arange(mask2.shape[0])
            # remove last row
            mask2[idx_row[np.logical_not(row_check)][-1], :] = True
            # remove first row
            mask2[idx_row[np.logical_not(row_check)][0], :] = True
            if self.par['inst_mode'] == 'LL':
               mask2[idx_row[np.logical_not(row_check)][:4], :] = True
               mask2[idx_row[np.logical_not(row_check)][-2:], :] = True

            mask = np.logical_and(mask1, mask2)
        return mask

    def _get_wavelength_calibration(self, numberOfWavelengthPixels, nodOffset):
        """
        Get wavelength calibration file.

        Parameters
        ----------
        numberOfWavelengthPixels : 'int'
        nodOffset : 'float'
        """
        wavelength_unit = u.micron

        wave_pixel_grid = np.arange(numberOfWavelengthPixels) * u.pix

        if self.par["obs_data"] == 'SPECTRUM':
            position_pixel_grid = np.zeros_like(wave_pixel_grid)
            spectral_trace = \
                collections.OrderedDict(wavelength_pixel=wave_pixel_grid,
                                        positional_pixel=position_pixel_grid,
                                        wavelength=self.data.wavelength.
                                        data[:, 0])
            return spectral_trace

        wave_cal_name = \
            os.path.join(self.par['obs_cal_path'],
                         self.par['inst_obs_name'],
                         self.par['inst_inst_name'],
                         self.par['obs_cal_version'],
                         'IRSX_'+self.par['inst_mode']+'_' +
                         self.par['obs_cal_version']+'_cal.wavsamp.tbl')
        wavesamp = ascii.read(wave_cal_name)
        order = wavesamp['order']

        spatial_pos = wavesamp['x_center']
        wavelength_pos = wavesamp['y_center']
        wavelength = wavesamp['wavelength']
        x0 = wavesamp['x0']
        x1 = wavesamp['x1']
        x2 = wavesamp['x2']
        x3 = wavesamp['x3']
        y0 = wavesamp['y0']
        y1 = wavesamp['y1']
        y2 = wavesamp['y2']
        y3 = wavesamp['y3']

        if self.par['inst_order'] == '1':
            idx = np.where(order == 1)
        elif (self.par['inst_order'] == '2') or \
                (self.par['inst_order'] == '3'):
            idx = np.where((order == 2) | (order == 3))

        spatial_pos = spatial_pos[idx].data - 0.5
        wavelength_pos = wavelength_pos[idx].data - 0.5
        wavelength = wavelength[idx].data
        x0 = x0[idx].data - 0.5
        x1 = x1[idx].data - 0.5
        x2 = x2[idx].data - 0.5
        x3 = x3[idx].data - 0.5
        y0 = y0[idx].data - 0.5
        y1 = y1[idx].data - 0.5
        y2 = y2[idx].data - 0.5
        y3 = y3[idx].data - 0.5

        spatial_pos_left = (x1+x2)/2
        spatial_pos_right = (x0+x3)/2
        wavelength_pos_left = (y1+y2)/2
        wavelength_pos_right = (y0+y3)/2

        f = interpolate.interp1d(wavelength_pos, spatial_pos,
                                 fill_value='extrapolate')
        spatial_pos_interpolated = f(wave_pixel_grid.value) * u.pix
        f = interpolate.interp1d(wavelength_pos, wavelength,
                                 fill_value='extrapolate')
        wavelength_interpolated = f(wave_pixel_grid.value) * wavelength_unit

        f = interpolate.interp1d(wavelength_pos_left, spatial_pos_left,
                                 fill_value='extrapolate')
        spatial_pos_left_interpolated = f(wave_pixel_grid.value) * u.pix
        f = interpolate.interp1d(wavelength_pos_left, wavelength,
                                 fill_value='extrapolate')
        wavelength_left_interpolated = \
            f(wave_pixel_grid.value) * wavelength_unit

        f = interpolate.interp1d(wavelength_pos_right, spatial_pos_right,
                                 fill_value='extrapolate')
        spatial_pos_right_interpolated = f(wave_pixel_grid.value) * u.pix
        f = interpolate.interp1d(wavelength_pos_right, wavelength,
                                 fill_value='extrapolate')
        wavelength_right_interpolated = \
            f(wave_pixel_grid.value) * wavelength_unit

        grid_x = np.hstack([spatial_pos_left_interpolated.value,
                            spatial_pos_interpolated.value,
                            spatial_pos_right_interpolated.value])
        grid_y = np.hstack([wave_pixel_grid.value,
                            wave_pixel_grid.value, wave_pixel_grid.value])
        points = np.array([grid_y, grid_x]).T
        values = np.hstack([wavelength_left_interpolated.value,
                            wavelength_interpolated.value,
                            wavelength_right_interpolated.value])

        corrected_spatial_pos = spatial_pos_interpolated+nodOffset
        wave_cal = griddata(points, values,
                            (wave_pixel_grid.value,
                             corrected_spatial_pos.value),
                            method='cubic') * wavelength_unit

        return wave_cal

    def get_detector_cubes(self, is_background=False):
        """
        Get 3D detector cubes.

        This function combines all functionallity to read fits files
        containing the (uncalibrated) detector cubes (detector data
        on ramp level) timeseries, including
        orbital phase and wavelength information

        Parameters
        ----------
        is_background : `bool`
            if `True` the data represents an observaton of the IR background
            to be subtracted of the data of the transit spectroscopy target.

        Returns
        -------
        SpectralTimeSeries : `cascade.data_model.SpectralDataTimeSeries`
            Instance of `SpectralDataTimeSeries` containing all spectroscopic
            data including uncertainties, time, wavelength and bad pixel mask.

        Raises
        ------
        AssertionError, KeyError
            Raises an error if no data is found or if certain expected
            fits keywords are not present in the data files.

        Notes
        -----
        Notes on timing in header:

        There are several integration-time-related keywords.
        Of greatest interest to the observer is the
        “effective integration time”, which is the time on-chip between
        the first and last non-destructive reads for each pixel. It is called:

            RAMPTIME = Total integration time for the current DCE.

        The value of RAMPTIME gives the usable portion of the integration ramp,
        occurring between the beginning of the first read and the end of the
        last read. It excludes detector array pre-conditioning time.
        It may also be of interest to know the exposure time at other points
        along the ramp. The SUR sequence consists of the time taken at the
        beginning of a SUR sequence to condition the array
        (header keyword DEADTIME), the time taken to complete one read and
        one spin through the array (GRPTIME), and the non-destructive reads
        separated by uniform wait times. The wait consists of “clocking”
        through the array without reading or resetting. The time it takes to
        clock through the array once is given by the SAMPTIME keyword.
        So, for an N-read ramp:

            RAMPTIME = 2x(N-1)xSAMPTIME

        and

           DCE duration = DEADTIME + GRPTIME + RAMPTIME

        Note that peak-up data is not obtained in SUR mode. It is obtained in
        Double Correlated Sampling (DCS) mode. In that case, RAMPTIME gives the
        time interval between the 2nd sample and the preceeding reset.

        """
        # order mask
        mask = self._get_order_mask()

        # get data files
        if is_background:
            obsid = self.par['obs_backgr_id']
            target_name = self.par['obs_backgr_target_name']
        else:
            obsid = self.par['obs_id']
            target_name = self.par['obs_target_name']
        if self.par['inst_mode'] == 'SL':
            path_to_files = self.par['obs_path'] + \
                target_name + '/IRSX/' + \
                self.par['obs_cal_version']+'/bcd/ch0/'
            data_files = find('SPITZER_S0*'+obsid +
                              '*lnz.fits', path_to_files)
        else:
            path_to_files = self.par['obs_path'] + \
                target_name + '/IRSX/' + \
                self.par['obs_cal_version']+'/bcd/ch2/'
            data_files = find('SPITZER_S2*'+obsid +
                              '*lnz.fits', path_to_files)

        # number of integrations
        nintegrations = len(data_files)
        if nintegrations < 2:
            raise AssertionError("No Timeseries data found in dir " +
                                 path_to_files)

        # get the size of the spectral images
        # In the spitzer IRS detector cubes, the first axis is
        # the number of frames. To get it in the proper data format
        # we need to move the frames axis to the back.
        image_file = data_files[0]
        # WARNING fits data is single precision!!
        spectral_image = np.moveaxis(fits.getdata(image_file, ext=0), 0, -1)
        # wavelength, spatial, frames axis
        npix, mpix, nframes = spectral_image.shape
        # get frametime from fits header (duration of ramp in sec)
        framtime = fits.getval(image_file, "FRAMTIME", ext=0)
        # get deadtime etc. from fits header
        # deadtime = fits.getval(image_file, "DEADTIME", ext=0)
        samptime = fits.getval(image_file, "SAMPTIME", ext=0)
        # get the FOVID from the header and check for nodded observations
        try:
            fovid = fits.getval(image_file, "FOVID", ext=0)
        except KeyError:
            print("FOVID not set in fits files")
            print("Using 'observations_mode' parameter instead")
            if self.par['obs_mode'] == "STARING":
                fovid = 28
            else:
                fovid = 26
        if (fovid != 28) and (fovid != 34) and (fovid != 40) and (fovid != 46):
            isNodded = True
            nodOffset = -5.0 * u.pix
        else:
            isNodded = False
            nodOffset = 0.0 * u.pix

        # get the unit of the spectral images
        try:
            flux_unit_string = fits.getval(image_file, "BUNIT", ext=0)
            flux_unit_string = flux_unit_string.replace("e-", "electron")
            flux_unit_string = flux_unit_string.replace("sec", "s")
            flux_unit = u.Unit(flux_unit_string)
        except KeyError:
            print("No flux unit set in fits files")
            flux_unit = u.dimensionless_unscaled

        # define mask and fill with data  from order mask
        mask = np.tile(mask.T, (nintegrations, nframes, 1, 1)).T

        # get the data
        # make sure time is last axis
        image_cube = np.zeros((npix, mpix, nframes, nintegrations))
        time = np.zeros((nintegrations))
        for im, image_file in enumerate(tqdm(data_files, dynamic_ncols=True)):
            # WARNING fits data is single precision!!
            spectral_image = \
                np.moveaxis(fits.getdata(image_file, ext=0), 0, -1)
            image_cube[:, :, :, im] = spectral_image
            time[im] = fits.getval(image_file, "BMJD_OBS", ext=0)

        image_cube = np.diff(image_cube, axis=2)
        mask = mask[:, :, :-1, :]
        flux_unit = flux_unit/(2.0*samptime*u.s)
        data = image_cube * flux_unit
        npix, mpix, nframes, nintegrations = data.shape

        # The time in the spitzer fits header is -2400000.5 and
        # it is the time at the start of the ramp
        # As we are using fitted ramps,
        # shift time by half ramp of length framtime
        time = (time + 2400000.5) + (0.50*framtime) / (24.0*3600.0)  # in days
        # orbital phase
        phase = (time - self.par['obj_ephemeris']) / self.par['obj_period']
        phase = phase - np.int(np.max(phase))
        if np.max(phase) < 0.0:
            phase = phase + 1.0

        # adjust time stamp for each sample up the ramp
        phase = np.tile(phase, (nframes, 1))
        time_shift = (np.arange(nframes)*2.0*samptime) - (0.50*framtime)
        time_shift = np.tile(time_shift, (nintegrations, 1)).T
        time_shift = time_shift / (24.0*3600.0) / self.par['obj_period']
        phase = phase + time_shift

        self._define_convolution_kernel()

        # wavelength calibration
        wave_cal = self._get_wavelength_calibration(npix, nodOffset)

        SpectralTimeSeries = SpectralDataTimeSeries(wavelength=wave_cal,
                                                    data=data, time=phase,
                                                    mask=mask,
                                                    isRampFitted=False,
                                                    isNodded=isNodded)
        return SpectralTimeSeries

    def get_spectral_trace(self):
        """Get spectral trace."""
        dim = self.data.data.shape
        wavelength_unit = self.data.wavelength_unit

        wave_pixel_grid = np.arange(dim[0]) * u.pix

        if self.par["obs_data"] == 'SPECTRUM':
            position_pixel_grid = np.zeros_like(wave_pixel_grid)
            spectral_trace = \
                collections.OrderedDict(wavelength_pixel=wave_pixel_grid,
                                        positional_pixel=position_pixel_grid,
                                        wavelength=self.data.wavelength.
                                        data[:, 0])
            return spectral_trace

        wave_cal_name = \
            os.path.join(self.par['obs_cal_path'],
                         self.par['inst_obs_name'],
                         self.par['inst_inst_name'],
                         self.par['obs_cal_version'],
                         'IRSX_'+self.par['inst_mode']+'_' +
                         self.par['obs_cal_version']+'_cal.wavsamp.tbl')
        wavesamp = ascii.read(wave_cal_name)
        order = wavesamp['order']

        isNodded = self.data.isNodded
        if isNodded:
            nodOffset = -5.0 * u.pix
        else:
            nodOffset = 0.0 * u.pix

        spatial_pos = wavesamp['x_center']
        wavelength_pos = wavesamp['y_center']
        wavelength = wavesamp['wavelength']
        x0 = wavesamp['x0']
        x1 = wavesamp['x1']
        x2 = wavesamp['x2']
        x3 = wavesamp['x3']
        y0 = wavesamp['y0']
        y1 = wavesamp['y1']
        y2 = wavesamp['y2']
        y3 = wavesamp['y3']

        if self.par['inst_order'] == '1':
            idx = np.where(order == 1)
        elif (self.par['inst_order'] == '2') or \
                (self.par['inst_order'] == '3'):
            idx = np.where((order == 2) | (order == 3))

        spatial_pos = spatial_pos[idx].data - 0.5
        wavelength_pos = wavelength_pos[idx].data - 0.5
        wavelength = wavelength[idx].data
        x0 = x0[idx].data - 0.5
        x1 = x1[idx].data - 0.5
        x2 = x2[idx].data - 0.5
        x3 = x3[idx].data - 0.5
        y0 = y0[idx].data - 0.5
        y1 = y1[idx].data - 0.5
        y2 = y2[idx].data - 0.5
        y3 = y3[idx].data - 0.5

        spatial_pos_left = (x1+x2)/2
        spatial_pos_right = (x0+x3)/2
        wavelength_pos_left = (y1+y2)/2
        wavelength_pos_right = (y0+y3)/2

        f = interpolate.interp1d(wavelength_pos, spatial_pos,
                                 fill_value='extrapolate')
        spatial_pos_interpolated = f(wave_pixel_grid.value) * u.pix
        f = interpolate.interp1d(wavelength_pos, wavelength,
                                 fill_value='extrapolate')
        wavelength_interpolated = f(wave_pixel_grid.value) * wavelength_unit

        f = interpolate.interp1d(wavelength_pos_left, spatial_pos_left,
                                 fill_value='extrapolate')
        spatial_pos_left_interpolated = f(wave_pixel_grid.value) * u.pix
        f = interpolate.interp1d(wavelength_pos_left, wavelength,
                                 fill_value='extrapolate')
        wavelength_left_interpolated = \
            f(wave_pixel_grid.value) * wavelength_unit

        f = interpolate.interp1d(wavelength_pos_right, spatial_pos_right,
                                 fill_value='extrapolate')
        spatial_pos_right_interpolated = f(wave_pixel_grid.value) * u.pix
        f = interpolate.interp1d(wavelength_pos_right, wavelength,
                                 fill_value='extrapolate')
        wavelength_right_interpolated = \
            f(wave_pixel_grid.value) * wavelength_unit

        grid_x = np.hstack([spatial_pos_left_interpolated.value,
                            spatial_pos_interpolated.value,
                            spatial_pos_right_interpolated.value])
        grid_y = np.hstack([wave_pixel_grid.value,
                            wave_pixel_grid.value, wave_pixel_grid.value])
        points = np.array([grid_y, grid_x]).T
        values = np.hstack([wavelength_left_interpolated.value,
                            wavelength_interpolated.value,
                            wavelength_right_interpolated.value])

        corrected_spatial_pos = spatial_pos_interpolated+nodOffset
        corrected_wavelengths = griddata(points, values,
                                         (wave_pixel_grid.value,
                                          corrected_spatial_pos.value),
                                         method='cubic') * wavelength_unit

        spectral_trace = collections.OrderedDict(
            wavelength_pixel=wave_pixel_grid,
            positional_pixel=corrected_spatial_pos[::-1],
            wavelength=corrected_wavelengths[::-1])

        return spectral_trace



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of the CASCADe package which has been
# developed within the ExoplANETS-A H2020 program.
#
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2020, 2021  Jeroen Bouwman
"""
Created on Mon May  4 19:08:58 2020

@author: Jeroen Bouwman
"""
# import copy
import os
import ast
import warnings
import numpy as np
import astropy.units as u
from astropy.visualization import quantity_support
from matplotlib import pyplot as plt
import seaborn as sns
from skimage import exposure
import statsmodels.distributions
import copy

from ..initialize import cascade_default_save_path
from ..initialize import cascade_configuration
from ..exoplanet_tools import transit_to_eclipse

__all__ = ["load_data_verbose",
           "subtract_background_verbose",
           "filter_dataset_verbose",
           "determine_source_movement_verbose",
           "check_wavelength_solution_verbose",
           "correct_wavelengths_verbose",
           "set_extraction_mask_verbose",
           "extract_1d_spectra_verbose",
           "calibrate_timeseries_verbose",
           "Verbose"]


def _get_plot_parameters():
    try:
        verbose = ast.literal_eval(cascade_configuration.cascade_verbose)
        save_verbose = \
            ast.literal_eval(cascade_configuration.cascade_save_verbose)
    except AttributeError:
        warnings.warn("Verbose flags not defined. Assuming False")
        verbose = False
        save_verbose = False
    try:
        save_path = cascade_configuration.cascade_save_path
        if not os.path.isabs(save_path):
            save_path = os.path.join(cascade_default_save_path, save_path)
        os.makedirs(save_path, exist_ok=True)
    except AttributeError:
        warnings.warn("No save path defined. Not saving plots")
        save_verbose = False
    try:
        observations_id = cascade_configuration.observations_id
        observations_target_name = \
            cascade_configuration.observations_target_name
        if observations_id not in observations_target_name:
            save_name_base = observations_target_name+'_'+observations_id
        else:
            save_name_base = observations_target_name
    except AttributeError:
        warnings.warn("No uniue id or target name set"
                      "save name not unique.")
        save_name_base = 'verbose'
    return (verbose, save_verbose, save_path, save_name_base)


def load_data_verbose(*args, **kwargs):
    """
    Make verbose plots for load_data step.

    Parameters
    ----------
    args : 'tuple'
        Input varables for plot output.
    kwargs : 'dict'
        Named varables for plot output. Included should be the following:
            verbose_par, plot_data

    Returns
    -------
    None.

    """
    (verbose, save_verbose, save_path, save_name_base) = kwargs["verbose_par"]
    if not verbose:
        return
    if "plot_data" not in kwargs.keys():
        return
    sns.set_context("talk", font_scale=1.5, rc={"lines.linewidth": 2.5})
    sns.set_style("white", {"xtick.bottom": True, "ytick.left": True})

    data = kwargs["plot_data"].dataset.return_masked_array("data")
    wavelength = kwargs["plot_data"].dataset.return_masked_array("wavelength")

    try:
        scaling = kwargs["plot_data"].dataset.return_masked_array("scaling")
        if scaling is None:
            scaling = 1.0
    except:
        scaling=1.0

    data=data/scaling

    fig, ax = plt.subplots(figsize=(12, 12), nrows=1, ncols=1)
    if data.ndim == 3:
        cmap = plt.cm.viridis
        cmap.set_bad("white", 1.)
        im_scaled = exposure.rescale_intensity(data[..., 0].filled(0.0))
        img_adapteq = exposure.equalize_adapthist(im_scaled, clip_limit=0.03)
        img_adapteq = np.ma.array(img_adapteq, mask=data[..., 0].mask)
        p = ax.imshow(img_adapteq,
                      origin="lower", aspect="auto",
                      cmap=cmap, interpolation="none", vmin=0.0, vmax=1.0)
        plt.colorbar(p, ax=ax).set_label("Normalized Intensity")
        ax.set_ylabel("Pixel Position Dispersion Direction")
        ax.set_xlabel("Pixel Position Cross-Dispersion Direction")
        fig_name_extension = "a"
    else:
        ax.plot(wavelength[:, 0], data[:, 0], lw=3, color="b")
        ax.set_ylabel("Siganl")
        ax.set_xlabel("Wavelength")
        fig_name_extension = "b"
    ax.set_title("First Integration {}.".format(save_name_base))
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_load_data_step_figure1{}.png".
                                 format(fig_name_extension)),
                    bbox_inches="tight")

    if not hasattr(kwargs["plot_data"].instrument_calibration,
                   "calibration_images"):
        return
    fig, ax = plt.subplots(figsize=(12, 12), nrows=1, ncols=1)
    cmap = plt.cm.viridis
    cmap.set_bad("white", 1.)
    image = \
        kwargs["plot_data"].instrument_calibration.calibration_images[0, ...]
    source_pos = kwargs["plot_data"].instrument_calibration.\
        calibration_source_position[0]
    expected_source_pos = kwargs["plot_data"].instrument_calibration.\
        expected_calibration_source_position[0]
    im_scaled = exposure.rescale_intensity(image)
    img_adapteq = exposure.equalize_adapthist(im_scaled, clip_limit=0.03)
    p = ax.imshow(img_adapteq,
                  origin="lower", aspect="auto",
                  cmap=cmap, interpolation="none", vmin=0.0, vmax=1.0)
    plt.colorbar(p, ax=ax).set_label("Normalized Intensity")
    ax.scatter(*source_pos, s=430,
               edgecolor="white", facecolor='none',
               label="Fitted position ({0:3.2f},{1:3.2f})".
               format(*source_pos))
    ax.scatter(*expected_source_pos, s=380,
               edgecolor="r", facecolor="none",
               label="Expected position ({0:3.2f},{1:3.2f})".
               format(*expected_source_pos))
    ax.set_title("Acquisition Image "
                 "Position {}".format(save_name_base))
    ax.legend()
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_load_data_step_figure2a.png"),
                    bbox_inches="tight")


def subtract_background_verbose(*args, **kwargs):
    """
    Make verbose plots for the subtract_background step.

    Parameters
    ----------
    args : 'tuple'
        Input varables for plot output.
    kwargs : 'dict'
        Named varables for plot output. Included should be the following:
            verbose_par, plot_data

    Returns
    -------
    None.

    """
    (verbose, save_verbose, save_path, save_name_base) = kwargs["verbose_par"]
    if not verbose:
        return
    if "plot_data" not in kwargs.keys():
        return
    sns.set_context("talk", font_scale=1.5, rc={"lines.linewidth": 2.5})
    sns.set_style("white", {"xtick.bottom": True, "ytick.left": True})

    data = kwargs["plot_data"].dataset.return_masked_array("data")
    wavelength = kwargs["plot_data"].dataset.return_masked_array("wavelength")
    time = kwargs["plot_data"].dataset.return_masked_array("time")
    roi = kwargs["plot_data"].instrument_calibration.roi

    try:
        scaling = kwargs["plot_data"].dataset.return_masked_array("scaling")
        if scaling is None:
            scaling = 1.0
    except:
        scaling=1.0

    data=data/scaling

    if data.ndim == 3:
        roi_cube = np.tile(roi.T, (time.shape[-1], 1, 1)).T
    else:
        roi_cube = np.tile(roi.T, (time.shape[-1], 1)).T
    data_with_roi = \
        np.ma.array(data,
                    mask=np.ma.mask_or(data.mask, roi_cube))
    wavelength_with_roi = \
        np.ma.array(wavelength,
                    mask=np.ma.mask_or(wavelength.mask, roi_cube))
    total_data = np.ma.sum(data_with_roi, axis=-1)/time.shape[-1]
    total_wavelength = np.ma.sum(wavelength_with_roi, axis=-1)/time.shape[-1]

    if data.ndim == 3:
        lightcurve = np.ma.sum(data_with_roi, axis=(0, 1))
        time = time[0, 0, :].data
    else:
        lightcurve = np.ma.sum(data_with_roi, axis=(0))
        time = time[0, :].data

    fig, ax = plt.subplots(figsize=(12, 12))
    if total_data.ndim == 2:
        cmap = plt.cm.viridis
        cmap.set_bad("white", 1.)
        im_scaled = exposure.rescale_intensity(total_data.filled(0.0))
        img_adapteq = exposure.equalize_adapthist(im_scaled, clip_limit=0.03)
        img_adapteq = np.ma.array(img_adapteq, mask=total_data.mask)
        p = ax.imshow(img_adapteq,
                      origin="lower", aspect="auto",
                      cmap=cmap, interpolation="none", vmin=0.0, vmax=1.0)
        plt.colorbar(p, ax=ax).set_label("Normalized Average Intensity")
        ax.set_ylabel("Pixel Position Dispersion Direction")
        ax.set_xlabel("Pixel Position Cross-Dispersion Direction")
        fig_name_extension = "a"
    else:
        ax.plot(total_wavelength, total_data)
        ax.set_xlabel("Wavelength")
        ax.set_ylabel("Average Signal")
        fig_name_extension = "b"
    ax.set_title("Background subtracted averaged "
                 "data {}.".format(save_name_base))
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_subtract_background_step_figure1{}.png".
                                 format(fig_name_extension)),
                    bbox_inches='tight')

    fig, ax = plt.subplots(figsize=(12, 12))
    ax.plot(time, lightcurve, ".")
    ax.set_xlabel("Orbital phase")
    ax.set_ylabel("Total Signal")
    ax.set_title("Background subtracted data {}.".format(save_name_base))
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_subtract_background_step_figure2{}.png".
                                 format(fig_name_extension)),
                    bbox_inches="tight")


def filter_dataset_verbose(*args, **kwargs):
    """
    Make verbose plots.

    Returns
    -------
    None.

    """
    pass


def determine_source_movement_verbose(*args, **kwargs):
    """
    Make verbose plots.

    Returns
    -------
    None.

    """
    pass


def correct_wavelengths_verbose(*args, **kwargs):
    """
    Make verbose plots.

    Returns
    -------
    None.

    """
    pass


def check_wavelength_solution_verbose(*args, **kwargs):
    """
    Make verbose plots for the check_wavelength_solution step.

    Parameters
    ----------
    args : 'tuple'
        Input varables for plot output.
    kwargs : 'dict'
        Named varables for plot output. Included should be the following:
            verbose_par, modeled_observations, corrected_observations

    Returns
    -------
    None.

    """
    (verbose, save_verbose, save_path, save_name_base) = kwargs["verbose_par"]
    if not verbose:
        return
    if "modeled_observations" not in kwargs.keys():
        return
    if "corrected_observations" not in kwargs.keys():
        return
    if "stellar_model" not in  kwargs.keys():
        return
    if "extension" not in  kwargs.keys():
        extension=""
    else:
        extension = kwargs["extension"]
    modeled_observations = kwargs["modeled_observations"]
    corrected_observations = kwargs["corrected_observations"]
    stellar_model =  kwargs["stellar_model"]

    sns.set_context("talk", font_scale=2.0, rc={"lines.linewidth": 5.5})
    sns.set_style("white", {"xtick.bottom": True, "ytick.left": True})

    wavelength_unit = modeled_observations[0].data.unit

    fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
    ax0 = axes
    ax0.plot(modeled_observations[0],
             modeled_observations[1]*modeled_observations[2] , label='Model')
    ax0.plot(corrected_observations[0],
             corrected_observations[1], label='Corrected Observations')
    ax0.plot(corrected_observations[0]-corrected_observations[2],
             corrected_observations[1], label='Un-Corrected Observations')
    plt.plot([], [], ' ',
             label="Used scaling: {:10.4f}".format(modeled_observations[2]))
    plt.plot([], [], ' ',
             label="Wavelength shift: {:10.4f}".format(corrected_observations[2]))
    ax0.set_xlabel('Wavelength [{}]'.format(wavelength_unit))
    ax0.set_ylabel('Normalized Signal')
    ax0.set_title("Comparison Model with Observed Mean Spectrum")
    ax0.legend(loc='best', fancybox=True, framealpha=1.0,
               ncol=1, bbox_to_anchor=(0.2, 0.10, 0.8, 0.8), shadow=True,
               handleheight=1.5, labelspacing=0.05, fontsize=20).set_zorder(11)
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_check_wavelength_solution" +
                                 extension + ".png"),
                    bbox_inches="tight")


def set_extraction_mask_verbose(*args, **kwargs):
    """
    Make verbose plots.

    Returns
    -------
    None.

    """
    pass


def extract_1d_spectra_verbose(*args, **kwargs):
    """
    Make verbose plots.

    Returns
    -------
    None.

    """
    pass


def calibrate_timeseries_verbose(*args, **kwargs):
    """
    Make verbose plots for the calibrate_timeserie step.

    Parameters
    ----------
    args : 'tuple'
        Input varables for plot output.
    kwargs : 'dict'
        Named varables for plot output. Included should be the following:
            verbose_par, model, dataset, cleaned_dataset, exoplanet_spectrum,
            calibration_results.

    Returns
    -------
    None.

    """
    (verbose, save_verbose, save_path, save_name_base) = kwargs["verbose_par"]
    if not verbose:
        return
    if "exoplanet_spectrum" not in kwargs.keys():
        return
    if "calibration_results" not in kwargs.keys():
        return
    if "model" not in kwargs.keys():
        return
    if "dataset" not in kwargs.keys():
        return
    if "cleaned_dataset" not in kwargs.keys():
        return
    if "stellar_modeling" not in kwargs.keys():
        has_stellar_model = False
    else:
        has_stellar_model = True
    exoplanet_spectrum = kwargs["exoplanet_spectrum"]
    calibration_results = kwargs["calibration_results"]
    model = kwargs["model"]
    dataset = kwargs["dataset"]
    cleaned_dataset = kwargs["cleaned_dataset"]
    if has_stellar_model:
        stellar_modeling = kwargs["stellar_modeling"]

    #######################################
    #  Fit quality and regularization
    ########################################
    sns.set_context("talk", font_scale=2.0, rc={"lines.linewidth": 4.5})
    sns.set_style("white", {"xtick.bottom": True, "ytick.left": True})

    wavelength_unit = exoplanet_spectrum.spectrum.wavelength_unit

    fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
    ax0 = axes
    ax0.plot(calibration_results.wavelength_normed_fitted_spectrum[0, :],
             np.log10(calibration_results.regularization))
    ax0.set_xlabel('Wavelength [{}]'.format(wavelength_unit))
    ax0.set_ylabel('log10 Alpha')
    ax0.set_title("Regularization Strength")
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_calibrate_timeseries_regularization.png"),
                    bbox_inches="tight")

    fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
    ax0 = axes
    ax0.plot(calibration_results.wavelength_normed_fitted_spectrum[0, :],
             np.mean(calibration_results.dof[1:, :], axis=0))
    ax0.set_xlabel('Wavelength [{}]'.format(wavelength_unit))
    ax0.set_ylabel('DOF')
    ax0.set_title("Degrees of Freedom")
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_calibrate_timeseries_dof.png"),
                    bbox_inches="tight")

    sigma_mse_cut = ast.literal_eval(cascade_configuration.cpm_sigma_mse_cut)
    fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
    ax0 = axes
    ax0.plot(calibration_results.wavelength_normed_fitted_spectrum[0, :],
             calibration_results.mse[1, :]*10000)
    plt.axhline(np.median(calibration_results.mse[1, :]*10000)*sigma_mse_cut,
                linestyle='dashed', color='black', label='rejection threshold')
    ax0.set_xlabel('Wavelength [{}]'.format(wavelength_unit))
    ax0.set_ylabel('MSE [x 10000]')
    ax0.set_title("MSE regression fit")
    ax0.legend(loc='lower left', fancybox=True, framealpha=1.0,
               ncol=1,
               bbox_to_anchor=(0, 0.90, 1, 0.2), shadow=True,
               handleheight=1.5, labelspacing=0.05,
               fontsize=20).set_zorder(11)
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_calibrate_timeseries_mse.png"),
                    bbox_inches="tight")

    fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
    ax0 = axes
    ax0.plot(calibration_results.wavelength_normed_fitted_spectrum[0, :],
             np.mean(calibration_results.aic[1:, :], axis=0))
    ax0.set_xlabel('Wavelength [{}]'.format(wavelength_unit))
    ax0.set_ylabel('AIC')
    ax0.set_title("AIC regression fit")
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_calibrate_timeseries_aic.png"),
                    bbox_inches="tight")
    ########################################
    # PDF
    ########################################
    from astropy.stats import knuth_bin_width
    from scipy.stats import loggamma
    sns.set_context("talk", font_scale=2.0, rc={"lines.linewidth": 4.5})
    sns.set_style("white", {"xtick.bottom": True, "ytick.left": True})

    transittype = model.transittype
    depth = (u.Quantity(cascade_configuration.object_radius) /
             u.Quantity(cascade_configuration.object_radius_host_star))
    depth = depth.decompose().value**2
    depth = depth*100
    nboot = int(cascade_configuration.cpm_nbootstrap)
    spectrum_data_unit = exoplanet_spectrum.spectrum_bootstrap.data_unit
    bad_wavelength_mask = exoplanet_spectrum.spectrum_bootstrap.mask
    bad_wavelength_mask = \
        np.repeat(bad_wavelength_mask[np.newaxis, :],
                  nboot+1,
                  axis=0)

    normed_spectrum = \
        np.ma.array(calibration_results.normed_fitted_spectra.copy(),
                    mask=bad_wavelength_mask.copy())
    if transittype == "secondary":
        normed_spectrum = transit_to_eclipse(normed_spectrum)
    normed_spectrum.data[...] = normed_spectrum.data*100
    mean_norm = np.ma.median(normed_spectrum[1:, :], axis=1)

    dx, bins = knuth_bin_width(mean_norm, return_bins=True)
    height, bin_edges = np.histogram(mean_norm, bins=bins, density=False)
    f = np.sum(height*np.diff(bins))
    distribution_function = loggamma
    distribution_variables = distribution_function.fit(mean_norm)
    TD_min_loggamma, TD_loggamma, TD_max_loggamma = \
        distribution_function(*distribution_variables).ppf([0.05, 0.5, 0.95])

    TD005, TD, TD095 = exoplanet_spectrum.spectrum_bootstrap.TDDEPTH

    x_kde = np.linspace(mean_norm.min(), mean_norm.max(), len(mean_norm))

    fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
    ax0 = axes
    _ = ax0.hist(mean_norm, bins=bins, density=False, alpha=0.5)
    ax0.plot(x_kde,
             distribution_function(*distribution_variables).pdf(x_kde)*f,
             label='log gamma')
    ax0.axvline(TD, linestyle='dashed', color='blue', label='Median Depth')
    ax0.fill_betweenx([0, height.max()], TD005, TD095, color='g', alpha=0.1,
                      label='95% confidence')
    ax0.set_title("PDF")
    if transittype == 'secondary':
        ax0.set_xlabel('Fplanet/Fstar [{}]'.format(spectrum_data_unit))
    else:
        plt.axvline(depth, linestyle='dashed', color='black')
        ax0.set_xlabel('Transit Depth [{}]'.format(spectrum_data_unit))
    ax0.set_ylabel('Number of bootstrap samples')
    ax0.legend(loc='lower left', fancybox=True, framealpha=1.0,
               ncol=1,
               bbox_to_anchor=(0, 0.90, 1, 0.2), shadow=True,
               handleheight=1.5, labelspacing=0.05,
               fontsize=20).set_zorder(11)
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_calibrate_timeseries_pdf.png"),
                    bbox_inches="tight")

    # ##################### CDF ###########################
    sns.set_context("talk", font_scale=2.0, rc={"lines.linewidth": 6.5})
    sns.set_style("white", {"xtick.bottom": True, "ytick.left": True})
    ecdf = statsmodels.distributions.ECDF(mean_norm)
    fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
    ax0 = axes
    ax0.plot(x_kde, ecdf(x_kde), label="Empirical CDF")
    ax0.plot(x_kde, distribution_function(*distribution_variables).cdf(x_kde),
             label="log gamma CDF")
    if transittype == 'secondary':
        ax0.set_xlabel('Fplanet/Fstar [{}]'.format(spectrum_data_unit))
    else:
        ax0.set_xlabel('Transit Depth [{}]'.format(spectrum_data_unit))
    ax0.set_title("CDF of Band Avaraged Transit Depth")
    ax0.set_ylabel('Cumulative Fraction')
    ax0.legend(loc='lower left', fancybox=True, framealpha=1.0,
               ncol=1,
               bbox_to_anchor=(0, 0.80, 1, 0.2), shadow=True,
               handleheight=1.5, labelspacing=0.05,
               fontsize=20).set_zorder(11)
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_calibrate_timeseries_cdf.png"),
                    bbox_inches="tight")

    # ############# QQ ######################
    sns.set_context("talk", font_scale=1.0, rc={"lines.linewidth": 3.5})
    sns.set_style("white", {"xtick.bottom": True, "ytick.left": True})
    from scipy.stats import probplot
    try:
        probplot(mean_norm, dist=distribution_function(*distribution_variables),
			            plot=plt.figure(dpi=200).add_subplot(111))
    except ValueError:
        pass
    plt.title("QQ-plot of mean transit depth")
    plt.show()
    if save_verbose:
        fig.savefig(os.path.join(save_path, save_name_base +
                                 "_calibrate_timeseries_qq.png"),
                    bbox_inches="tight")

    #################################
    #  Spectrum
    #################################
    sns.set_context("talk", font_scale=2.0, rc={"lines.linewidth": 3.5})
    sns.set_style("white", {"xtick.bottom": True, "ytick.left": True})

    transittype = model.transittype
    spectrum0 = exoplanet_spectrum.spectrum.return_masked_array('data')
    wave0 = exoplanet_spectrum.spectrum.return_masked_array('wavelength')
    error0 = exoplanet_spectrum.spectrum.return_masked_array('uncertainty')
    spectrum_data_unit = exoplanet_spectrum.spectrum.data_unit
    wavelength_unit = exoplanet_spectrum.spectrum.wavelength_unit
    spectrum_boot = \
        exoplanet_spectrum.spectrum_bootstrap.return_masked_array('data')
    wave_boot = \
        exoplanet_spectrum.spectrum_bootstrap.return_masked_array('wavelength')
    error_boot = \
        exoplanet_spectrum.spectrum_bootstrap.return_masked_array('uncertainty')

    TD005, TD, TD095 = exoplanet_spectrum.spectrum_bootstrap.TDDEPTH

    fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
    ax0 = axes
    ax0.errorbar(wave_boot, spectrum_boot, yerr=error_boot,
                 fmt=".", color='brown', lw=5, alpha=0.9, ecolor='brown',
                 markerfacecolor='brown', label='Bootstrap',
                 markeredgecolor='brown', fillstyle='full', markersize=15)
    ax0.errorbar(wave0, spectrum0, yerr=error0,
                 fmt=".", color='blue', lw=5, alpha=0.9, ecolor='blue',
                 markerfacecolor='blue', label='Single fit',
                 markeredgecolor='blue', fillstyle='full', markersize=15)
    plt.axhline(TD, linestyle='dashed', color='black')
    plt.fill_between([wave0.data[0], wave0.data[-1]],
                     TD005, TD095, color='g', alpha=0.1)
    if transittype == 'secondary':
        ax0.axes.set_ylim([0, 2.5*TD])
        ax0.axes.set_xlim([wave0.data[0], wave0.data[-1]])
        #ax0.set_ylabel('Fplanet/Fstar [{}]'.format(spectrum_data_unit))
        ax0.set_ylabel(f'F$_\mathrm{{p}}$ / F$_\mathrm{{s}}$ '
                       f'[{spectrum_data_unit.to_string(format="latex")}]')
    else:
        ax0.axes.set_ylim([TD*0.9, TD*1.1])
        ax0.axes.set_xlim([wave0.data[0], wave0.data[-1]])
        #ax0.set_ylabel('Transit Depth [{}]'.format(spectrum_data_unit))
        ax0.set_ylabel(f'R$^2_\mathrm{{p}}$ / R$^2_\mathrm{{s}}$ '
                       f'[{spectrum_data_unit.to_string(format="latex")}]')
    ax0.set_xlabel(f'Wavelength [{wavelength_unit.to_string(format="latex")}]')
    ax0.legend(loc='lower left', fancybox=True, framealpha=1.0,
               ncol=1,
               bbox_to_anchor=(0, 0.90, 1, 0.2), shadow=True,
               handleheight=1.5, labelspacing=0.05,
               fontsize=20).set_zorder(11)
    plt.show()
    if save_verbose:
        fig.savefig(
            os.path.join(save_path, save_name_base +
                         "_calibrate_timeseries_exoplanet_spectrum.png"),
            bbox_inches="tight")

    if transittype == "secondary":
        with quantity_support():
            mask = copy.deepcopy(exoplanet_spectrum.brightness_temperature.mask)
            mask += ~np.isfinite(exoplanet_spectrum.brightness_temperature.
                                 uncertainty.data.value)
            fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
            ax0 = axes
            ax0.errorbar(exoplanet_spectrum.brightness_temperature.
                         wavelength.data[~mask].to(u.cm**-1,
                                                   equivalencies=u.spectral()),
                         exoplanet_spectrum.brightness_temperature.
                         data.data[~mask],
                         yerr=exoplanet_spectrum.brightness_temperature.
                         uncertainty.data[~mask],
                         fmt=".", color='blue', lw=5, alpha=0.9, ecolor='blue',
                         markerfacecolor='blue', markeredgecolor='blue',
                         fillstyle='full', markersize=20
                         )
            ax0.set_ylabel(f'Brightness Temperature [{ax0.yaxis.units.to_string(format="latex")}]')
            ax0.set_xlabel(f'Wavelength [{ax0.xaxis.units.to_string(format="latex")}]')
            plt.show()
        if save_verbose:
            fig.savefig(
                os.path.join(save_path, save_name_base +
                             "_calibrate_timeseries_brightness_temperature_spectrum.png"),
                bbox_inches="tight")

    ####################################
    # residual plot
    ##################################
    sns.set_context("talk", font_scale=1.3, rc={"lines.linewidth": 3.5})
    sns.set_style("white", {"xtick.bottom": True, "ytick.left": True})

    residual_unit = u.percent
    image_res = calibration_results.fitted_residuals_bootstrap.data * 100
    wave0 = calibration_results.fitted_residuals_bootstrap.wavelength
    wavelength_unit = calibration_results.fitted_residuals_bootstrap.wavelength_unit

    wave0_min = np.ma.min(wave0).data.value
    wave0_max = np.ma.max(wave0).data.value
    fig, ax = plt.subplots(figsize=(7, 6), dpi=200)
    for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                 ax.get_xticklabels() + ax.get_yticklabels()):
        item.set_fontsize(20)
    p = ax.imshow(image_res,
                  origin='lower',
                  cmap='hot',
                  interpolation='nearest',
                  aspect='auto',
                  extent=[0, image_res.shape[1], wave0_min, wave0_max])
    plt.colorbar(p, ax=ax).set_label("Residual ({})".format(residual_unit))
    ax.set_xlabel("Integration Number")
    ax.set_ylabel('Wavelength [{}]'.format(wavelength_unit))
    ax.set_title('Residual Image')
    plt.show()
    if save_verbose:
        fig.savefig(
            os.path.join(save_path, save_name_base +
                         "_calibrate_timeseries_fit_residual.png"),
            bbox_inches="tight")

    ############################################################
    # Systematics model
    ############################################################
    sns.set_context("talk", font_scale=2.0, rc={"lines.linewidth": 5.5})
    sns.set_style("white", {"xtick.bottom": True, "ytick.left": True})

    TD005, TD, TD095 = exoplanet_spectrum.spectrum_bootstrap.TDDEPTH
    systematics_model = calibration_results.fitted_systematics_bootstrap
    systematics = systematics_model.return_masked_array('data')
    time_systematics = systematics_model.return_masked_array('time')

    roi_cube = cleaned_dataset.data.mask
    uncal_data = \
        dataset.return_masked_array('data').copy()[~np.all(roi_cube,
                                                           axis=1), ...]
    uncal_data.mask = np.ma.logical_or(uncal_data.mask, systematics.mask)
    cal_data = uncal_data/systematics
# Bug Fix
    TD_temp = TD/100  # *np.mean(model.limbdarkning_correction)
    if transittype == 'secondary':
        from cascade.exoplanet_tools import eclipse_to_transit
        TD_temp = eclipse_to_transit(TD_temp)
    model_lc = np.mean(model.light_curve_interpolated /
                       model.dilution_correction*TD_temp, axis=0) + 1

    fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
    ax0 = axes
    ax0.plot(np.ma.mean(time_systematics, axis=0),
             np.ma.mean(systematics, axis=0), '.',
             markersize=20, label='Bootstraped Systematics Model')
    ax0.set_xlabel('Orbital Phase')
    ax0.set_ylabel('Mean Signal [{}]'.format(systematics_model.data_unit))
    ax0.set_title("Band Averaged Instrument Systematics")
    ax0.legend(loc='lower left', fancybox=True, framealpha=1.0,
               ncol=1,
               bbox_to_anchor=(0, 0.23, 1, 0.2), shadow=True,
               handleheight=1.5, labelspacing=0.05,
               fontsize=20).set_zorder(11)
    plt.show()
    if save_verbose:
        fig.savefig(
            os.path.join(save_path, save_name_base +
                         "_calibrate_timeseries_systematics.png"),
            bbox_inches="tight")

    fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
    ax0 = axes
    ax0.plot(np.ma.mean(time_systematics, axis=0),
             np.ma.mean(cal_data[:, :], axis=0), '.',
             zorder=6, markersize=25, label='Calibrated Data')
    ax0.plot(np.ma.mean(time_systematics, axis=0),
             np.ma.mean(uncal_data, axis=0)/np.ma.mean(systematics),
             '.', zorder=5, markersize=25, label='Uncalibrated Data')
    ax0.plot(np.ma.mean(time_systematics, axis=0), model_lc, '.',
             zorder=8, markersize=20, label='Fitted Lightcurve Model')
    ax0.axes.set_ylim([1-1.3*TD/100, 1+0.8*TD/100])
    ax0.set_xlabel('Orbital Phase')
    ax0.set_ylabel('Mean Signal [{}]'.format(systematics_model.data_unit))
    ax0.set_title("Band Averaged Calibrated Timeseries")
    ax0.legend(loc='lower left', fancybox=True, framealpha=1.0,
               ncol=1,
               bbox_to_anchor=(0, 0.83, 1, 0.2), shadow=True,
               handleheight=1.5, labelspacing=0.05,
               fontsize=20).set_zorder(11)
    plt.show()
    if save_verbose:
        fig.savefig(
            os.path.join(save_path, save_name_base +
                         "_calibrate_timeseries_calibrated_lightcurve.png"),
            bbox_inches="tight")

    #########################################################################
    # Stellar Spetrum
    #########################################################################
    if has_stellar_model:
        with quantity_support():
            scaling = exoplanet_spectrum.flux_calibrated_stellar_model.SCALING
            mask = copy.deepcopy(exoplanet_spectrum.flux_calibrated_stellar_spectrum.mask)
            fig, axes = plt.subplots(figsize=(18, 12), nrows=1, ncols=1, dpi=200)
            ax0 = axes
            ax0.errorbar(exoplanet_spectrum.flux_calibrated_stellar_spectrum.
                         wavelength.data[~mask],
                         exoplanet_spectrum.flux_calibrated_stellar_spectrum.
                         data.data[~mask],
                         yerr=exoplanet_spectrum.flux_calibrated_stellar_spectrum.
                         uncertainty.data[~mask],
                         fmt=".", color='brown', lw=5, alpha=0.9, ecolor='brown',
                         markerfacecolor='brown', markeredgecolor='brown',
                         fillstyle='full', markersize=20, zorder=7,
                         label="Observed Stellar Spectrum"
                         )
            ax0.plot(exoplanet_spectrum.flux_calibrated_stellar_model.
                     wavelength.data[~mask],
                     exoplanet_spectrum.flux_calibrated_stellar_model.
                     data.data[~mask]*scaling, label="Stellar Model",
                     color="b", lw=8, zorder=8)
            plt.plot([], [], ' ',
                     label="Used scaling: {:10.4f}".format(scaling))

            ax0.set_ylabel(f'Flux [{ax0.yaxis.units.to_string(format="latex")}]')
            ax0.set_xlabel(f'Wavelength [{ax0.xaxis.units.to_string(format="latex")}]')
            ax0.set_title("Comparison Model with Observed Stellar Spectrum")
            ax0.legend(loc='lower left', fancybox=True, framealpha=1.0,
                       ncol=1, bbox_to_anchor=(0.1, 0.80, 1, 0.3), shadow=True,
                       handleheight=1.5, labelspacing=0.05,
                       fontsize=20).set_zorder(11)
            plt.show()
        if save_verbose:
            fig.savefig(
                os.path.join(save_path, save_name_base +
                             "_calibrate_timeseries_calibrated_stellar_spectrum.png"),
                bbox_inches="tight")

    ####################3
    # Check distribution
    ######################
    from scipy import stats
    sns.set_context("talk", font_scale=1.0, rc={"lines.linewidth": 2.5})
    sns.set_style("white", {"xtick.bottom": True, "ytick.left": True})

    def normal(mean, std, histmax=False, color="black", label=''):
        x = np.linspace(mean-4*std, mean+4*std, 200)
        p = stats.norm.pdf(x, mean, std)
        if histmax:
            p = p*histmax/max(p)
        z = plt.plot(x, p, color, linewidth=2, label=label)

    spectrum_boot = \
        exoplanet_spectrum.spectrum_bootstrap.return_masked_array('data')
    wave_boot = \
        exoplanet_spectrum.spectrum_bootstrap.return_masked_array('wavelength')
    error_boot = \
        exoplanet_spectrum.spectrum_bootstrap.return_masked_array('uncertainty')

    indx = (wave_boot > np.ma.min(wave_boot)*1.05) & (wave_boot < np.ma.max(wave_boot)*0.95)
    data = spectrum_boot[indx]*1.e4
    median_data = np.ma.median(data)
    data -= median_data
    error = error_boot[indx]*1.e4

    av_std = []
    for i in range(300):
        av_std.append(np.std(np.random.normal(loc=0.0, scale=error)))
    av_std = np.mean(av_std)

    fig, ax = plt.subplots(figsize=(8, 5), ncols=1, nrows=1, dpi=200)
    his = sns.histplot(x=data, stat="probability", ax=ax)
    normal(data.mean(), data.std(), histmax=ax.get_ylim()[1],
           label=f'Fit to distribution, $\sigma$={data.std(): .2f}')
    normal(0.0, av_std, histmax=ax.get_ylim()[1], color='red',
           label=f'Expected distribution from errors, $\sigma$={av_std: .2f}')
    ax.set_xlabel('Deviation from median depth [ppm]')
    ax.legend(loc='upper left', fancybox=False, framealpha=0.5,
                  ncol=1, bbox_to_anchor=(0.0, 0.01, 1, 1), shadow=False,
                  handleheight=1.4, labelspacing=0.1,
                  fontsize=10).set_zorder(11)
    plt.show()
    if save_verbose:
        fig.savefig(
            os.path.join(save_path, save_name_base +
                         "_distribution_from_median_depth.png"),
            bbox_inches="tight")


    data = np.diff(spectrum_boot[indx]*1.e4)
    median_data = np.median(data)
    data -= median_data

    av_std = []
    for i in range(300):
        av_std.append(np.std(np.diff(
            np.random.normal(loc=0.0, scale=error))))
    av_std = np.mean(av_std)

    fig, ax = plt.subplots(figsize=(8, 5), ncols=1, nrows=1, dpi=200)
    his = sns.histplot(x=data, stat="probability", ax=ax)
    normal(data.mean(), data.std(), histmax=ax.get_ylim()[1],
           label=f'Fit to distribution, $\sigma$={data.std(): .2f}')
    normal(0.0, av_std, histmax=ax.get_ylim()[1], color='red',
           label=f'Expected distribution from errors, $\sigma$={av_std: .2f}')
    ax.set_xlabel('Pairwise difference [ppm]')
    ax.legend(loc='upper left', fancybox=False, framealpha=0.5,
                  ncol=1, bbox_to_anchor=(0.0, 0.01, 1, 1), shadow=False,
                  handleheight=1.4, labelspacing=0.1,
                  fontsize=10).set_zorder(11)
    plt.show()
    if save_verbose:
        fig.savefig(
            os.path.join(save_path, save_name_base +
                         "_pairwise_differences_distribution.png"),
            bbox_inches="tight")



class Verbose:
    """The Class handels verbose output vor the cascade pipeline."""

    def __init__(self):
        self.verbose_par = _get_plot_parameters()

    @property
    def __valid_commands(self):
        """
        All valid pipeline commands.

        This function returns a dictionary with all the valid commands
        which can be parsed to the instance of the TSO object.
        """
        return {"load_data": load_data_verbose,
                "subtract_background": subtract_background_verbose,
                "filter_dataset": filter_dataset_verbose,
                "determine_source_movement": determine_source_movement_verbose,
                "correct_wavelengths": correct_wavelengths_verbose,
                "check_wavelength_solution": check_wavelength_solution_verbose,
                "set_extraction_mask": set_extraction_mask_verbose,
                "extract_1d_spectra": extract_1d_spectra_verbose,
                "calibrate_timeseries": calibrate_timeseries_verbose,
                }

    def execute(self, command, *args, **kwargs):
        """
        Excecute the pipeline commands.

        This function checks if a command is valid and excecute it if True.

        Parameters
        ----------
        command : `str`
            Command to be excecuted. If valid the method corresponding
            to the command will be excecuted

        Raises
        ------
        ValueError
            error is raised if command is not valid

        Examples
        --------
        Example how to run the command to reset a tso object:

        >>> vrbs.execute('load_data')

        """
        if command not in self.__valid_commands:
            raise ValueError("Command not recognized, "
                             "check your data reduction command for the "
                             "following valid commands: {}. Aborting "
                             "command".format(self.__valid_commands.keys()))

        self.__valid_commands[command](*args, **kwargs,
                                       verbose_par=self.verbose_par)
